"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This script contains all functions related to the random loose packing (RLP).
Script version date: 23.11.2024
"""  

""" Import packages """
import math
import random
import numpy as np
from functools import reduce
from scipy.optimize import fsolve
from scipy.optimize import minimize

""" Functions """
def calc_position(config, z, y, x, z_drop, y_drop, x_drop, d, M_particle_steady, test_range, d_max_steady, particle_i):
    """ 
    Function / Algorithm to determine the steady state position of the considered particle within the RLP.
    In this loop it is checked whether a particle sediments or collides with other particles.
    If the considered particle does not overlap with other particles, it is moved downwards.
    If the considered particle overlaps with other particles, 
    depending on the number of overlappings, its position is updated by the corresponding function.
    The position is updated so that the particles touch each other but do not overlap.
    After the position update the considered particleis tested for overlapping again because it could 
    overlap with other particles due to the update. 
    Depending on the number of overlaps, the position is updated over and over until no overlapping after 
    the position update occurs. Only when no overlappings are present it is moved downwards.
    When three or four overlappings are present, the particle is checked for steady state.
    If it is not in steady state, it is calculated at which particle it would roll along in the next loop step 
    and the corresponding position update function is applied. This is neccessary to prevent the particle to get stucked.
    This loop is repeated agian and again until the considered particle reches its steady state position at the ground, 
    at three or at four particles.
    """
    # Counter variables and lists
    # Number of overplaps
    n_overlaps = 0
    # List of overlaps
    overlaps = []
    # Variable for overlapping walls
    walls = 0
    # Error Variables
    error_position_update2 = 0
    error_position_update3 = 0
    error_position_update1w = 0
    error_position_update2w = 0
    error_position_update1w2 = 0
    error_position_update2w2 = 0
    error_check_steadystate3 = 0
    error_check_steadystate2w = 0
    error_check_steadystate1w2 = 0
    error_check_steadystate2w2 = 0
    # Argument for the while loop
    steady = 0
    # Aborting variable for the while loop 
    counter_sedi = 0
    # Borders of the aborting variable for the while loop
    counter_sedi_max1 = 10000
    counter_sedi_max2 = 2 * counter_sedi_max1
    counter_sedi_max3 = 3 * counter_sedi_max1
    # Error counter
    n_error = 0
    # If fatal error is True, the while loop is leaved
    fatal_error = False
    # While loop fpr particle sedimentation
    while steady == 0:
        # Increase aborting variable for the while loop
        counter_sedi = counter_sedi + 1
        # Try again if it takes too long or an error occurs
        if counter_sedi == counter_sedi_max1 or n_error == 1:
            z = z_drop
            y = y_drop
            x = x_drop
            if n_error == 1:
                n_error = n_error + 1
            continue
        # Assign new coordinates if it takes again too long or an error occurs again
        if counter_sedi == counter_sedi_max2 or n_error == 3:
            x = random.uniform(-config.A_0 + d / 2, config.A_0 - d / 2)
            y = random.uniform(-config.A_0 + d / 2, config.A_0 - d / 2)
            z = find_z_drop(y, x, M_particle_steady, test_range, d_max_steady)
            if n_error == 3:
                n_error = n_error + 1
            continue
        # Abort while loop by abort variable if trying again and assigning new coordinates does not work 
        # or an error occurs three times
        if counter_sedi > counter_sedi_max3 or n_error == 5:
            n_error = 0
            if config.display_detailed_progress == True:
                print('Timeout in the sedimentation loop or too many errors. Check sedimentation loop.')
            fatal_error = True
            break
        # Calculate overlapping/collision
        n_overlaps, overlaps, walls = calc_overlap(config, d, z, y, x, M_particle_steady, test_range)
        # If the considered particle does not overlap with other particles, it is moved downwards.
        if n_overlaps == 0:
            # If z == d/2, the considered particle reached the ground
            if z == d/2:
                # If steady = 1, steady state position is reched on the ground
                steady = 1
            else:
                z = z - config.stepsize   
        # Overlap with one steady state particle in total
        elif n_overlaps == 1:
            # index of the first steady state particle, the considered particle collides with
            index1 = overlaps[0]
            # No overlappings with walls
            if walls == 0:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with one steady state particle.')
                # If the considered particle overlaps with one steady state particle, 
                # its position is updated by position_update1 function. 
                y, x = position_update1(y, x, d, M_particle_steady, overlaps)
            # Overlap with one wall   
            elif walls == 1 or walls == -1 or walls == 2 or walls == -2:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with one steady state particle and wall.')
                # If the considered particle overlaps with one steady state particle and one wall, 
                # its position is updated by position_update1w function.
                y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                # If no range error occurs, the new coordinates are assignet
                if error_position_update1w == 0:
                    y = y_s
                    x = x_s
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                elif error_position_update1w == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                    y = y_s
                    x = x_s                    
                # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                elif error_position_update1w == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                    z = z + config.stepsize / 2
                    continue
                # Error value is 3 if the both solver solutions lies behind the simulation borders
                elif error_position_update1w == 3:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                    n_error = n_error + 1
                    continue
            # Overlap with two walls
            elif walls == 3 or walls == -3 or walls == 4 or walls == -4:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with one steady state particle and two walls.')
                # If the considered particle overlaps with one steady state particle and two walls, 
                # its position is updated by position_update1w function.
                z_s, y_s, x_s, error_position_update1w2 = position_update1w2(config, z, y, x, d, M_particle_steady, overlaps, walls)
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                if error_position_update1w2 == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                # Error value is 2 if the position of the considered particle is located below the centroid of the steady state particles
                elif error_position_update1w2 == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w2 function. Position of the considered particle is located below the centroid of the steady state particles.')
                    z = z + config.stepsize / 2
                    continue
                # Error value is 3 if the position of the first solver solution (the logical one) is not located on the intersection plane
                elif error_position_update1w2 == 3:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w2 function. The logical new position determnined by the solver is not located on the intersection plane. ')
                    z = z + config.stepsize / 2
                    continue
                # If particleis in contact with one steady state particle and two walls, it could be in steady state
                # Checking for overlaps after position update
                n_overlaps, overlaps1w2, walls = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                # If no overlaps occure, the particle is tested for steady state
                if n_overlaps == 0:
                    # With three contact partners, the considered particle can be in a steady state, 
                    # which is tested by the check_steadystate1w2 function
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'is tested for steady state at one particle and two walls.')
                    steady_check, overlaps, walls, error_check_steadystate1w2 = check_steadystate1w2(config, z_s, y_s, x_s, M_particle_steady, overlaps, walls)
                    # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                    if error_check_steadystate1w2 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                        n_error = n_error + 1
                        continue
                    # Error value is 2 if the triangle points of the steady state particles are collinear
                    elif error_check_steadystate1w2 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                        n_error = n_error + 1
                        continue
                    # Error value is 3 if the subdomain of the triangle could not be determined correctly
                    elif error_check_steadystate1w2 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                        n_error = n_error + 1
                        continue
                    # If steady_check == 0, the particle is in steady state
                    if steady_check == 3:
                        # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                        z = z_s
                        y = y_s
                        x = x_s
                        # If steady = 2, steady state position is reched on one particle and two walls
                        steady = 6
                    # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                    # Thus its position is updated by position_update1 function
                    elif steady_check == 1:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                        # Function for position update at one contact.
                        y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                    # If steady_check == 2, the particle is not in steady state and would roll along one steady state particle and wall
                    # Thus its position is updated by position_update2 function    
                    elif steady_check == 4:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle and wall')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update1w == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update1w == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update1w == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update1w == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                    continue
                # If overlaps occur, the considered particle is moved 1/10 of the step upwards an handled again in the next loop.
                else:        
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles and two walls.')
                    z = z + config.stepsize / 2
                    continue
            # Checking for overlaps after position update
            n_overlaps, overlaps, walls = calc_overlap(config, d, z, y, x, M_particle_steady, test_range)
            # Overlap with two steady state particles in total
            if n_overlaps == 1:
                # index of the second steady state particle, the considered particle collides with
                index2 = overlaps[0]
                # Update overlaps array
                overlaps = np.array([index1, index2])
                # No overlappings with walls
                if walls == 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with two steady state particles.')
                    # If the considered particle overlaps with one steady state particle after position update at one particle,
                    # it is in contact with two particles. Thus, its position is updated by position_update2 function.
                    y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                    # If no range error occurs, the new coordinates are assignet
                    if error_position_update2 == 0:
                        y = y_s
                        x = x_s
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    elif error_position_update2 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                        y = y_s
                        x = x_s                    
                    # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                    # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                    elif error_position_update2 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                        z = z + config.stepsize / 2
                        continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                    elif error_position_update2 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                        n_error = n_error + 1
                        continue
                # Overlap with one wall
                elif walls == 1 or walls == -1 or walls == 2 or walls == -2:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with two steady state particles and one wall.')
                    # If the considered particle overlaps with one steady state particle and one wall after position update at one particle,
                    # it is in contact with two particles and one wall. Thus, its position is updated by position_update2w function.
                    z_s, y_s, x_s, error_position_update2w = position_update2w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    if error_position_update2w == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                    # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                    # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                    elif error_position_update2w == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                        z = z + config.stepsize / 2
                        continue
                    # If particle is in contact with two steady state particle and one wall, it could be in steady state
                    # Checking for overlaps after position update
                    n_overlaps, overlaps2w, walls = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                    # If the particle overlaps with a second wall, its position is updated in the next loop
                    if walls == 3 or walls == -3 or walls == 4 or walls == -4 or walls == 0:
                        continue
                    # If no overlaps occure, the particle is tested for steady state
                    if n_overlaps == 0:
                        # With three contact partners, the considered particle can be in a steady state, 
                        # which is tested by the check_steadystate1w2 function
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'is tested for steady state at two particles and one wall.')
                        steady_check, overlaps, error_check_steadystate2w = check_steadystate2w(config, z_s, y_s, x_s, M_particle_steady, overlaps, walls)
                        # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                        if error_check_steadystate2w == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                            n_error = n_error + 1
                            continue
                        # Error value is 2 if the triangle points of the steady state particles are collinear
                        elif error_check_steadystate2w == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                            n_error = n_error + 1
                            continue
                        # Error value is 3 if the subdomain of the triangle could not be determined correctly
                        elif error_check_steadystate2w == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                            n_error = n_error + 1
                            continue
                        # If steady_check == 0, the particle is in steady state
                        if steady_check == 3:
                            # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                            z = z_s
                            y = y_s
                            x = x_s
                            # If steady = 2, steady state position is reched at two particles and one wall
                            steady = 4
                        # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                        # Thus its position is updated by position_update1 function
                        elif steady_check == 1:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                            # Function for position update at one contact.
                            y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                        # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                        # Thus its position is updated by position_update2 function                          
                        elif steady_check == 2:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                            # Function for position update at two contacts.
                            y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                            # If no range error occurs, the new coordinates are assignet
                            if error_position_update2 == 0:
                                y = y_s
                                x = x_s
                            # Error value is 1 if the solver found the same new position of the considered particle twice.
                            # The error is only displayed but the calculation is continued, as the position could be correct
                            elif error_position_update2 == 1:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                                y = y_s
                                x = x_s                    
                            # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                            # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                            elif error_position_update2 == 2:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                                z = z + config.stepsize / 2
                                continue
                            # Error value is 3 if the both solver solutions lies behind the simulation borders
                            elif error_position_update2 == 3:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                                n_error = n_error + 1
                                continue
                        # If steady_check == 4, the particle is not in steady state and would roll along one steady state particle and wall
                        # Thus its position is updated by position_update1w function    
                        elif steady_check == 4:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle and wall')
                            # Function for position update at two contacts.
                            y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                            # If no range error occurs, the new coordinates are assignet
                            if error_position_update1w == 0:
                                y = y_s
                                x = x_s
                            # Error value is 1 if the solver found the same new position of the considered particle twice.
                            # The error is only displayed but the calculation is continued, as the position could be correct
                            elif error_position_update1w == 1:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                                y = y_s
                                x = x_s                    
                            # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                            # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                            elif error_position_update1w == 2:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                                z = z + config.stepsize / 2
                                continue
                            # Error value is 3 if the both solver solutions lies behind the simulation borders
                            elif error_position_update1w == 3:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                                n_error = n_error + 1
                                continue
                        continue
                    # If overlaps occur, the considered particle is moved 1/10 of the step upwards an handled again in the next loop.
                    else:        
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'collides with three steady state particles and two walls.')
                        z = z + config.stepsize / 2
                        continue
                # Overlap with two walls
                elif walls == 3 or walls == -3 or walls == 4 or walls == -4:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with one steady state particle and two walls.')
                    # If the considered particle overlaps with one steady state particle and two walls after position update at one particle,
                    # it is in contact with two particles and two walls. Thus, its position is updated by position_update2w function.
                    z_s, y_s, x_s, error_position_update2w2 = position_update2w2(config, z, y, x, d, M_particle_steady, overlaps, walls)
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    if error_position_update2w2 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2w2 function. Solver found the same new position twice. Calculation proceeds with the found value.')                  
                    # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                    # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                    elif error_position_update2w2 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2w2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                        z = z + config.stepsize / 2
                        continue
                    # If particle is in contact with two steady state particle and two walls, it could be in steady state
                    # Checking for overlaps after position update
                    n_overlaps, overlaps2w2, walls = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                    # If the particle overlaps with onlyone wall, its position is updated in the next loop
                    if walls == 1 or walls == -1 or walls == 2 or walls == -2 or walls == 0:
                        continue
                    # If no overlaps occure, the particle is tested for steady state
                    if n_overlaps == 0:
                        # With three contact partners, the considered particle can be in a steady state, 
                        # which is tested by the check_steadystate1w2 function
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'is tested for steady state at two particles and two walls.')
                        steady_check, overlaps, walls, error_check_steadystate2w2 = check_steadystate2w2(config, z_s, y_s, x_s, M_particle_steady, overlaps, walls)
                        # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                        if error_check_steadystate2w2 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                            n_error = n_error + 1
                            continue
                        # Error value is 2 if the triangle points of the steady state particles are collinear
                        elif error_check_steadystate2w2 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                            n_error = n_error + 1
                            continue
                        # Error value is 3 if the subdomain of the triangle could not be determined correctly
                        elif error_check_steadystate2w2 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                            n_error = n_error + 1
                            continue
                        # If steady_check == 0, the particle is in steady state
                        if steady_check == 3:
                            # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                            z = z_s
                            y = y_s
                            x = x_s
                            # If steady = 2, steady state position is reched at two particles and two walls
                            steady = 7
                        # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                        # Thus its position is updated by position_update1 function
                        elif steady_check == 1:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                            # Function for position update at one contact.
                            y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                        # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                        # Thus its position is updated by position_update2 function                          
                        elif steady_check == 2:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                            # Function for position update at two contacts.
                            y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                            # If no range error occurs, the new coordinates are assignet
                            if error_position_update2 == 0:
                                y = y_s
                                x = x_s
                            # Error value is 1 if the solver found the same new position of the considered particle twice.
                            # The error is only displayed but the calculation is continued, as the position could be correct
                            elif error_position_update2 == 1:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                                y = y_s
                                x = x_s                    
                            # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                            # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                            elif error_position_update2 == 2:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                                z = z + config.stepsize / 2
                                continue
                            # Error value is 3 if the both solver solutions lies behind the simulation borders
                            elif error_position_update2 == 3:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                                n_error = n_error + 1
                                continue
                        # If steady_check == 4, the particle is not in steady state and would roll along one steady state particle and wall
                        # Thus its position is updated by position_update1w function    
                        elif steady_check == 4:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle and wall')
                            # Function for position update at two contacts.
                            y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                            # If no range error occurs, the new coordinates are assignet
                            if error_position_update1w == 0:
                                y = y_s
                                x = x_s
                            # Error value is 1 if the solver found the same new position of the considered particle twice.
                            # The error is only displayed but the calculation is continued, as the position could be correct
                            elif error_position_update1w == 1:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                                y = y_s
                                x = x_s                    
                            # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                            # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                            elif error_position_update1w == 2:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                                z = z + config.stepsize / 2
                                continue
                            # Error value is 3 if the both solver solutions lies behind the simulation borders
                            elif error_position_update1w == 3:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                                n_error = n_error + 1
                                continue
                        continue
                    # If overlaps occur, the considered particle is moved 1/10 of the step upwards an handled again in the next loop.
                    else:        
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'collides with three steady state particles and two walls.')
                        z = z + config.stepsize / 2
                        continue
                # Checking for overlaps after position update
                n_overlaps, overlaps, walls = calc_overlap(config, d, z, y, x, M_particle_steady, test_range)
                # Three overlappings in total
                if n_overlaps == 1:
                    # index of the third steady state particle, the considered particle collides with
                    index3 = overlaps[0]
                    # Update overlaps array
                    overlaps = np.array([index1, index2, index3])
                    # No overlappings with walls
                    if walls == 0:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'collides with three steady state particles.')
                        # If the considered particle overlaps with one steady state particle after position update at two particles,
                        # it is in contact with three particles. Thus, its position is updated by position_update3 function.
                        z_s, y_s, x_s, error_position_update3 = position_update3(config, z, y, x, d, M_particle_steady, overlaps)
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        if error_position_update3 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update3 function. Solver found the same new position twice.  Calculation proceeds with the found value.')
                        # Error value is 2 if the position of the considered particle is located below the centroid of the steady state particles
                        elif error_position_update3 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update3 function. Position of the considered particle is located below the centroid of the steady state particles.')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the position of the first solver solution (the logical one) is not located on the intersection plane
                        elif error_position_update3 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update3 function. The logical new position determnined by the solver is not located on the intersection plane. ')
                            z = z + config.stepsize / 2
                            continue
                    # If the considered particle overlaps with more than one wall after position update at three particles, 
                    # it is moved 1/10 of the step upwards an handled again in the next loop.
                    else:        
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'collides with three steady state particles and walls.')
                        z = z + config.stepsize / 2
                        continue
                    # Checking for overlaps after position update 
                    n_overlaps, overlaps3, error_overlap = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                    # If value is True, an error has occurred in the calc_overlapping function.
                    # The considered particle is stored in the steady state array with NaN values.
                    if error_overlap == True:
                        if config.display_detailed_progress == True:
                            print('Error in the calc_overlapping function. Check calc_overlapping function.')
                        fatal_error = True
                        break
                    # If no overlaps occure, the particle is tested for steady state
                    if n_overlaps == 0:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i ,'is checked for steady state at three steady state particles.')
                        # With three contact partners, the considered particle can be in a steady state, 
                        # which is tested by the check_steadystate3 function
                        steady_check, overlaps, error_check_steadystate3 = check_steadystate3(z_s, y_s, x_s, M_particle_steady, overlaps)
                        # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                        if error_check_steadystate3 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                            n_error = n_error + 1
                            continue
                        # Error value is 2 if the triangle points of the steady state particles are collinear
                        elif error_check_steadystate3 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                            n_error = n_error + 1
                            continue
                        # Error value is 3 if the subdomain of the triangle could not be determined correctly
                        elif error_check_steadystate3 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                            n_error = n_error + 1
                            continue
                        # If steady_check == 0, the particle is in steady state
                        if steady_check == 3:
                            # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                            z = z_s
                            y = y_s
                            x = x_s
                            # If steady = 2, steady state position is reched on three particles
                            steady = 2
                        # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                        # Thus its position is updated by position_update1 function
                        elif steady_check == 1:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                            # Function for position update at one contact.
                            y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                        # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                        # Thus its position is updated by position_update2 function    
                        elif steady_check == 2:
                            if config.display_detailed_progress == True:
                                print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                            # Function for position update at two contacts.
                            y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                            # If no range error occurs, the new coordinates are assignet
                            if error_position_update2 == 0:
                                y = y_s
                                x = x_s
                            # Error value is 1 if the solver found the same new position of the considered particle twice.
                            # The error is only displayed but the calculation is continued, as the position could be correct
                            elif error_position_update2 == 1:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                                y = y_s
                                x = x_s                    
                            # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                            # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                            elif error_position_update2 == 2:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                                z = z + config.stepsize / 2
                                continue
                            # Error value is 3 if the both solver solutions lies behind the simulation borders
                            elif error_position_update2 == 3:
                                if config.display_detailed_progress == True:
                                    print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                                n_error = n_error + 1
                                continue
                    # If the considered particle overlaps with more than one steady state particle after position update at three contacts,
                    # it is in contact with more than four particles. 
                    # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
                    elif n_overlaps > 0:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'collides with more than three particles')
                        z = z + config.stepsize / 2
                        continue
                # If the considered particle overlaps with more than two steady state particle after position update at two contacts,
                # it is in contact with more than four particles. 
                # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
                elif n_overlaps > 1:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i, 'collides with more than three particles')
                    z = z + config.stepsize / 2
                    continue
            # Overlap with three steady state particles in total
            elif n_overlaps == 2:
                # Indices of the second and third steady state particles, the considered particle collides with
                index2 = overlaps[0]
                index3 = overlaps[1]
                # Update overlaps array
                overlaps = np.array([index1, index2, index3])
                # No overlappings with walls
                if walls == 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles.')
                    # If the considered particle overlaps with two steady state particle after position update at one contact,
                    # it is in contact with three particles. Thus, its position is updated by position_update3 function.
                    z_s, y_s, x_s, error_position_update3 = position_update3(config, z, y, x, d, M_particle_steady, overlaps)
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    if error_position_update3 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. Solver found the same new position twice.  Calculation proceeds with the found value.')
                    # Error value is 2 if the position of the considered particle is located below the centroid of the steady state particles
                    elif error_position_update3 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. Position of the considered particle is located below the centroid of the steady state particles.')
                        z = z + config.stepsize / 2
                        continue
                    # Error value is 3 if the position of the first solver solution (the logical one) is not located on the intersection plane
                    elif error_position_update3 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. The logical new position determnined by the solver is not located on the intersection plane. ')
                        z = z + config.stepsize / 2
                        continue
                # If the considered particle overlaps with more than one wall after position update at three particles, 
                # it is moved 1/10 of the step upwards an handled again in the next loop.
                else:        
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles and walls.')
                    z = z + config.stepsize / 2
                    continue
                # Checking for overlaps after position update   
                n_overlaps, overlaps3, error_overlap = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                # If value is True, an error has occurred in the calc_overlapping function.
                # The considered particle is stored in the steady state array with NaN values.
                if error_overlap == True:
                    if config.display_detailed_progress == True:
                        print('Error in the calc_overlapping function. Check calc_overlapping function.')
                    fatal_error = True
                    break
                # If no overlaps occure, the particle is tested for steady state
                if n_overlaps == 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'is checked for steady state at three steady state particles.')
                    # With three contact partners, the considered particle can be in a steady state, 
                    # which is tested by the check_steadystate3 function
                    steady_check, overlaps, error_check_steadystate3 = check_steadystate3(z_s, y_s, x_s, M_particle_steady, overlaps)
                    # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                    if error_check_steadystate3 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                        n_error = n_error + 1
                        continue
                    # Error value is 2 if the triangle points of the steady state particles are collinear
                    elif error_check_steadystate3 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                        n_error = n_error + 1
                        continue
                    # Error value is 3 if the subdomain of the triangle could not be determined correctly
                    elif error_check_steadystate3 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                        n_error = n_error + 1
                        continue
                    # If steady_check == 0, the particle is in steady state
                    if steady_check == 3:
                        # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                        z = z_s
                        y = y_s
                        x = x_s
                        # If steady = 2, steady state position is reched on three particles
                        steady = 2
                    # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                    # Thus its position is updated by position_update1 function
                    elif steady_check == 1:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                        # Function for position update at one contact.
                        y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                    # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                    # Thus its position is updated by position_update2 function    
                    elif steady_check == 2:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update2 == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update2 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update2 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update2 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                # If the considered particle overlaps with more than one steady state particle after position update at three contacts,
                # it is in contact with more than four particles. 
                # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
                elif n_overlaps > 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i, 'collides with more than three particles')
                    z = z + config.stepsize / 2
                    continue
            # If the considered particle overlaps with more than three steady state particle after position update at one contact,
            # it is in contact with more than four particles. 
            # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
            elif n_overlaps > 2:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i, 'collides with more than three particles')
                z = z + config.stepsize / 2    
                continue
        # Two overlaps with steady state particles in total
        elif n_overlaps == 2:
            # Indices of the second and third steady state particles, the considered particle collides with
            index1 = overlaps[0]
            index2 = overlaps[1]
            # No overlappings with walls
            if walls == 0:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with two steady state particles.')
                # If the considered particle overlaps with two steady state particles, 
                # its position is updated by position_update2 function.
                y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                # If no range error occurs, the new coordinates are assignet
                if error_position_update2 == 0:
                    y = y_s
                    x = x_s
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                elif error_position_update2 == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                    y = y_s
                    x = x_s                    
                # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                elif error_position_update2 == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                    z = z + config.stepsize / 2
                    continue
                # Error value is 3 if the both solver solutions lies behind the simulation borders
                elif error_position_update2 == 3:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                    n_error = n_error + 1
                    continue
            # Overlap with one wall
            elif walls == 1 or walls == -1 or walls == 2 or walls == -2:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with two steady state particles and one wall.')
                # If the considered particle overlaps with one steady state particle and one wall after position update at one particle,
                # it is in contact with two particles and one wall. Thus, its position is updated by position_update2w function.
                z_s, y_s, x_s, error_position_update2w = position_update2w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                if error_position_update2w == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2w function. Solver found the same new position twice. Calculation proceeds with the found value.')                 
                # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                elif error_position_update2w == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                    z = z + config.stepsize / 2
                    continue
                # If particle is in contact with two steady state particle and one wall, it could be in steady state
                # Checking for overlaps after position update
                n_overlaps, overlaps2w, walls = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                # If the particle overlaps with a second wall, its position is updated in the next loop
                if walls == 3 or walls == -3 or walls == 4 or walls == -4 or walls == 0:
                    continue
                # If no overlaps occure, the particle is tested for steady state
                if n_overlaps == 0:
                    # With three contact partners, the considered particle can be in a steady state, 
                    # which is tested by the check_steadystate1w2 function
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'is tested for steady state at two particles and one wall.')
                    steady_check, overlaps, error_check_steadystate2w = check_steadystate2w(config, z_s, y_s, x_s, M_particle_steady, overlaps, walls)
                    # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                    if error_check_steadystate2w == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                        n_error = n_error + 1
                        continue
                    # Error value is 2 if the triangle points of the steady state particles are collinear
                    elif error_check_steadystate2w == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                        n_error = n_error + 1
                        continue
                    # Error value is 3 if the subdomain of the triangle could not be determined correctly
                    elif error_check_steadystate2w == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                        n_error = n_error + 1
                        continue
                    # If steady_check == 0, the particle is in steady state
                    if steady_check == 3:
                        # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                        z = z_s
                        y = y_s
                        x = x_s
                        # If steady = 2, steady state position is reched at two particles and one wall
                        steady = 4
                    # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                    # Thus its position is updated by position_update1 function
                    elif steady_check == 1:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                        # Function for position update at one contact.
                        y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                    # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                    # Thus its position is updated by position_update2 function                          
                    elif steady_check == 2:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update2 == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update2 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update2 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update2 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                    # If steady_check == 4, the particle is not in steady state and would roll along one steady state particle and wall
                    # Thus its position is updated by position_update1w function    
                    elif steady_check == 4:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle and wall')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update1w == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update1w == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update1w == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update1w == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                    continue
                # If overlaps occur, the considered particle is moved 1/10 of the step upwards an handled again in the next loop.
                else:        
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles and two walls.')
                    z = z + config.stepsize / 2
                    continue
            # Overlap with two walls
            elif walls == 3 or walls == -3 or walls == 4 or walls == -4:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with one steady state particle and two walls.')
                # If the considered particle overlaps with one steady state particle and two walls after position update at one particle,
                # it is in contact with two particles and two walls. Thus, its position is updated by position_update2w function.
                z_s, y_s, x_s, error_position_update2w2 = position_update2w2(config, z, y, x, d, M_particle_steady, overlaps, walls)
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                if error_position_update2w2 == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2w2 function. Solver found the same new position twice. Calculation proceeds with the found value.')                  
                # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                elif error_position_update2w2 == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update2w2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                    z = z + config.stepsize / 2
                    continue
                # If particle is in contact with two steady state particle and two walls, it could be in steady state
                # Checking for overlaps after position update
                n_overlaps, overlaps2w2, walls = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                # If the particle overlaps with onlyone wall, its position is updated in the next loop
                if walls == 1 or walls == -1 or walls == 2 or walls == -2 or walls == 0:
                    continue
                # If no overlaps occure, the particle is tested for steady state
                if n_overlaps == 0:
                    # With three contact partners, the considered particle can be in a steady state, 
                    # which is tested by the check_steadystate1w2 function
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'is tested for steady state at two particles and two walls.')
                    steady_check, overlaps, walls, error_check_steadystate2w2 = check_steadystate2w2(config, z_s, y_s, x_s, M_particle_steady, overlaps, walls)
                    # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                    if error_check_steadystate2w2 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                        n_error = n_error + 1
                        continue
                    # Error value is 2 if the triangle points of the steady state particles are collinear
                    elif error_check_steadystate2w2 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                        n_error = n_error + 1
                        continue
                    # Error value is 3 if the subdomain of the triangle could not be determined correctly
                    elif error_check_steadystate2w2 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                        n_error = n_error + 1
                        continue
                    # If steady_check == 0, the particle is in steady state
                    if steady_check == 3:
                        # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                        z = z_s
                        y = y_s
                        x = x_s
                        # If steady = 2, steady state position is reched at two particles and two walls
                        steady = 7
                    # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                    # Thus its position is updated by position_update1 function
                    elif steady_check == 1:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                        # Function for position update at one contact.
                        y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                    # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                    # Thus its position is updated by position_update2 function                          
                    elif steady_check == 2:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update2 == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update2 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update2 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update2 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                    # If steady_check == 4, the particle is not in steady state and would roll along one steady state particle and wall
                    # Thus its position is updated by position_update1w function    
                    elif steady_check == 4:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle and wall')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update1w = position_update1w(config, z, y, x, d, M_particle_steady, overlaps, walls)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update1w == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update1w == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update1w == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update1w == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                    continue
                # If overlaps occur, the considered particle is moved 1/10 of the step upwards an handled again in the next loop.
                else:        
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles and two walls.')
                    z = z + config.stepsize / 2
                    continue
            # Checking for overlaps after position update
            n_overlaps, overlaps, walls = calc_overlap(config, d, z, y, x, M_particle_steady, test_range)
            # Three overlaps with steady state particles in total
            if n_overlaps == 1:
                # Index of the second and third steady state particles, the considered particle collides with
                index3 = overlaps[0]
                # Update overlaps array
                overlaps = np.array([index1, index2, index3])
                # No overlappings with walls    
                if walls == 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles.')
                    # If the considered particle overlaps with one steady state particle after position update at two contacts,
                    # it is in contact with three particles. Thus, its position is updated by position_update3 function.
                    z_s, y_s, x_s, error_position_update3 = position_update3(config, z, y, x, d, M_particle_steady, overlaps)
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    if error_position_update3 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. Solver found the same new position twice.  Calculation proceeds with the found value.')
                    # Error value is 2 if the position of the considered particle is located below the centroid of the steady state particles
                    elif error_position_update3 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. Position of the considered particle is located below the centroid of the steady state particles.')
                        z = z + config.stepsize / 2
                        continue
                    # Error value is 3 if the position of the first solver solution (the logical one) is not located on the intersection plane
                    elif error_position_update3 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update3 function. The logical new position determnined by the solver is not located on the intersection plane. ')
                        z = z + config.stepsize / 2
                        continue
                # If the considered particle overlaps with more than one wall after position update at three particles, 
                # it is moved 1/10 of the step upwards an handled again in the next loop.
                else:        
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'collides with three steady state particles and walls.')
                    z = z + config.stepsize / 2
                    continue
                # Checking for overlaps after position update   
                n_overlaps, overlaps3, error_overlap = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
                # If value is True, an error has occurred in the calc_overlapping function.
                # The considered particle is stored in the steady state array with NaN values.
                if error_overlap == True:
                    if config.display_detailed_progress == True:
                        print('Error in the calc_overlapping function. Check calc_overlapping function.')
                    fatal_error = True
                    break
                # If no overlaps occure, the particle is tested for steady state
                if n_overlaps == 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i ,'is checked for steady state at three steady state particles.')
                    # With three contact partners, the considered particle can be in a steady state, 
                    # which is tested by the check_steadystate3 function
                    steady_check, overlaps, error_check_steadystate3 = check_steadystate3(z_s, y_s, x_s, M_particle_steady, overlaps)
                    # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                    if error_check_steadystate3 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                        n_error = n_error + 1
                        continue
                    # Error value is 2 if the triangle points of the steady state particles are collinear
                    elif error_check_steadystate3 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                        n_error = n_error + 1
                        continue
                    # Error value is 3 if the subdomain of the triangle could not be determined correctly
                    elif error_check_steadystate3 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                        n_error = n_error + 1
                        continue
                    # If steady_check == 0, the particle is in steady state
                    if steady_check == 3:
                        # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                        z = z_s
                        y = y_s
                        x = x_s
                        # If steady = 2, steady state position is reched on three particles
                        steady = 2
                    # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                    # Thus its position is updated by position_update1 function
                    elif steady_check == 1:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                        # Function for position update at one contact.
                        y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                    # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                    # Thus its position is updated by position_update2 function    
                    elif steady_check == 2:
                        if config.display_detailed_progress == True:
                            print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                        # Function for position update at two contacts.
                        y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                        # If no range error occurs, the new coordinates are assignet
                        if error_position_update2 == 0:
                            y = y_s
                            x = x_s
                        # Error value is 1 if the solver found the same new position of the considered particle twice.
                        # The error is only displayed but the calculation is continued, as the position could be correct
                        elif error_position_update2 == 1:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                            y = y_s
                            x = x_s                    
                        # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                        # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                        elif error_position_update2 == 2:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                            z = z + config.stepsize / 2
                            continue
                        # Error value is 3 if the both solver solutions lies behind the simulation borders
                        elif error_position_update2 == 3:
                            if config.display_detailed_progress == True:
                                print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                            n_error = n_error + 1
                            continue
                # If the considered particle overlaps with more than one steady state particle after position update at three contacts,
                # it is in contact with more than four particles. 
                # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
                elif n_overlaps > 0:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i, 'collides with more than three particles')
                    z = z + config.stepsize / 2
                    continue
            # If the considered particle overlaps with more than two steady state particle after position update at two contacts,
            # it is in contact with more than four particles. 
            # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
            elif n_overlaps > 1:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i, 'collides with more than four particles')
                z = z + config.stepsize / 2
                continue
        # Three overlaps with steady state particles in total
        elif n_overlaps == 3:
            # Indices of the first, second and third steady state particles, the considered particle collides with
            index1 = overlaps[0]
            index2 = overlaps[1]
            index3 = overlaps[2]
            # No overlappings with walls
            if walls == 0:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with three steady state particles.')
                # If the considered particle overlaps with three steady state particles, 
                # its position is updated by position_update3 function.
                z_s, y_s, x_s, error_position_update3 = position_update3(config, z, y, x, d, M_particle_steady, overlaps)
                # Error value is 1 if the solver found the same new position of the considered particle twice.
                # The error is only displayed but the calculation is continued, as the position could be correct
                if error_position_update3 == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update3 function. Solver found the same new position twice.  Calculation proceeds with the found value.')
                # Error value is 2 if the position of the considered particle is located below the centroid of the steady state particles
                elif error_position_update3 == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update3 function. Position of the considered particle is located below the centroid of the steady state particles.')
                    z = z + config.stepsize / 2
                    continue
                # Error value is 3 if the position of the first solver solution (the logical one) is not located on the intersection plane
                elif error_position_update3 == 3:
                    if config.display_detailed_progress == True:
                        print('Error in the error_position_update3 function. The logical new position determnined by the solver is not located on the intersection plane. ')
                    z = z + config.stepsize / 2
                    continue
            # If the considered particle overlaps with more than one wall after position update at three particles, 
            # it is moved 1/10 of the step upwards an handled again in the next loop.
            else:        
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'collides with three steady state particles and walls.')
                z = z + config.stepsize / 2
                continue
            # Checking for overlaps after position update   
            n_overlaps, overlaps3, error_overlap = calc_overlap(config, d, z_s, y_s, x_s, M_particle_steady, test_range)
            # If no overlaps occure, the particle is tested for steady state
            if n_overlaps == 0:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i ,'is checked for steady state at three steady state particles.')
                # With three contact partners, the considered particle can be in a steady state, 
                # which is tested by the check_steadystate3 function
                steady_check, overlaps, error_check_steadystate3 = check_steadystate3(z_s, y_s, x_s, M_particle_steady, overlaps)
                # Error value is 1 if particles center is located below the triangle's centroid of the steady state particles
                if error_check_steadystate3 == 1:
                    if config.display_detailed_progress == True:
                        print('Error in the check_steadystate3 function. The center of the considered particle is located below the triangles centroid of the steady state particles.')
                    n_error = n_error + 1
                    continue
                # Error value is 2 if the triangle points of the steady state particles are collinear
                elif error_check_steadystate3 == 2:
                    if config.display_detailed_progress == True:
                        print('Error in the check_steadystate3 function. The triangle points of the steady state particles are collinear.')
                    n_error = n_error + 1
                    continue
                # Error value is 3 if the subdomain of the triangle could not be determined correctly
                elif error_check_steadystate3 == 3:
                    if config.display_detailed_progress == True:
                        print('Error in the check_steadystate3 function. The subdomain of the triangle could not be determined correctly.')
                    n_error = n_error + 1
                    continue
                # If steady_check == 0, the particle is in steady state
                if steady_check == 3:
                    # Assigning the coordinates of the considered partcile to its steady state coordinates at three particles
                    z = z_s
                    y = y_s
                    x = x_s
                    # If steady = 2, steady state position is reched on three particles
                    steady = 2
                # If steady_check == 1, the particle is not in steady state and would roll along one steady state particle
                # Thus its position is updated by position_update1 function    
                elif steady_check == 1:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along one particle')
                    # Function for position update at one contact.
                    y, x = position_update1(y, x, d, M_particle_steady, overlaps)
                # If steady_check == 2, the particle is not in steady state and would roll along two steady state particles
                # Thus its position is updated by position_update2 function
                elif steady_check == 2:
                    if config.display_detailed_progress == True:
                        print('Particle Nr.', particle_i, 'exits the steady state check and is now rolling along two particles')
                    # Function for position update at two contacts.
                    y_s, x_s, error_position_update2 = position_update2(config, z, y, x, d, M_particle_steady, overlaps)
                    # If no range error occurs, the new coordinates are assignet
                    if error_position_update2 == 0:
                        y = y_s
                        x = x_s
                    # Error value is 1 if the solver found the same new position of the considered particle twice.
                    # The error is only displayed but the calculation is continued, as the position could be correct
                    elif error_position_update2 == 1:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2 function. Solver found the same new position twice. Calculation proceeds with the found value.')
                        y = y_s
                        x = x_s                    
                    # Error value is 2 if the distance between the centers of the two steady state particles and the position 
                    # of the first solver solution (the logical one) are not in the same range as the sums of their radii
                    elif error_position_update2 == 2:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update2 function. The logical solution of the new position determined by the solver lies not in the correct range. ')
                        z = z + config.stepsize / 2
                        continue
                    # Error value is 3 if the both solver solutions lies behind the simulation borders
                    elif error_position_update2 == 3:
                        if config.display_detailed_progress == True:
                            print('Error in the error_position_update1w function. The both solver solutions lies behind the simulation borders. ')
                        n_error = n_error + 1
                        continue
            # If the considered particle overlaps with more than one steady state particle after position update at three contacts,
            # it is in contact with more than four particles. 
            # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
            elif n_overlaps > 0:
                if config.display_detailed_progress == True:
                    print('Particle Nr.', particle_i, 'collides with more than four particles')
                z = z + config.stepsize / 2
                continue
        # If the considered particle overlaps with more than four steady state particle,
        # it is in contact with more than four particles. 
        # In this case, the particle is moved 1/10 of the step upwards an handled again in the next loop.
        elif n_overlaps > 3:
            if config.display_detailed_progress == True:
                print('Particle Nr.', particle_i, 'collides with more than four particles')
            z = z + config.stepsize / 2
            continue
        # If particle reaches the ground, it is stored in steady state array.
        if z <= d / 2:
            z = d / 2
    
    return z, y, x, steady, overlaps, fatal_error




def find_z_drop(y, x, M_particle_steady, test_range, d_max_steady):
    """ This function determines the drop position of particles """
    # Define testing area
    y_min = y - test_range
    y_max = y + test_range
    x_min = x - test_range
    x_max = x + test_range
    # Check particles in steady state for interraction with testing area:
    # in y direction
    particle_to_y = np.argwhere(np.logical_and(M_particle_steady[:, 2] > y_min, M_particle_steady[:, 2] < y_max))
    # in x direction
    particle_to_x = np.argwhere(np.logical_and(M_particle_steady[:, 3] > x_min, M_particle_steady[:, 3] < x_max))
    # Combine the results of both tests (arrays with indices of particles which interract with testing area)
    particle_to_check = reduce(np.intersect1d, (particle_to_y, particle_to_x))
    # Convert array to list
    particle_to_check = np.array(particle_to_check).tolist()
    # Create list with z-coordinates of interracting particles
    z_list = []
    for i in particle_to_check:
        value = M_particle_steady[i, 1]
        z_list.append(value)
    # Check list for entries and pick the max value
    if not z_list:
        max_z_value = 0
    else:
        max_z_value = max(z_list)
    # Calculate dropping height
    z = (max_z_value + d_max_steady)*1.2

    return z

def calc_overlap(config, d_i, z_i, y_i, x_i, M_particle_steady, test_range):
    """ 
    Function that checks overlaps of an unstable particle.
    It also double check the collision criteria of the calc_sedimentation function.
    If calc_sedimentation function works properly, only one overlapping should be identified.
    """
    # Variable to identifiycollisions with walls
    walls = 0 
    # Counting variable for number of collisions with steady state particles
    n_overlaps = 0
    # Array with indices of particles in steady state which overlap with the considered particle
    overlaps = []
    """ Define region of interrest """
    z_min = z_i - test_range
    if z_min < 0:
        z_min = 0
    z_max = z_i + test_range
    y_min = y_i - test_range
    y_max = y_i + test_range
    x_min = x_i - test_range
    x_max = x_i + test_range
    """ Checking particles within the testing range for possible overlappings """
    # Arrays in x, y and z direction with indices of particles, which could overlap with the considered particle
    x_array = np.argwhere(np.logical_and(M_particle_steady[:, 3] > x_min, M_particle_steady[:, 3] < x_max))
    y_array = np.argwhere(np.logical_and(M_particle_steady[:, 2] > y_min, M_particle_steady[:, 2] < y_max))
    z_array = np.argwhere(np.logical_and(M_particle_steady[:, 1] > z_min, M_particle_steady[:, 1] < z_max))
    # Combine the arrays
    particle_to_check = reduce(np.intersect1d, (x_array, y_array, z_array))
    # Convert array to list
    particle_to_check = np.array(particle_to_check).tolist()
    """ Check each particle that may overlap for an actual overlapping """
    for i in particle_to_check:
        # Retrieve information about particles that may overlap with the particle under consideration from the steady state array
        # Coordinates
        z_j = float(M_particle_steady[i, 1])
        y_j = float(M_particle_steady[i, 2])
        x_j = float(M_particle_steady[i, 3])
        # Diameter
        d_j = float(M_particle_steady[i, 5])
        # distance between the particle's centers when they collide
        dist_required = (d_i + d_j) / 2
        # Actual distance between the particle's centers 
        distance_particles = math.sqrt((z_i-z_j)**2 + (y_i-y_j)**2 + (x_i-x_j)**2)
        # Check whether particles overlap and store the overlapping particle in array
        if dist_required - distance_particles > 1e-9:
            overlaps.append(i)
    # Remove duplicates from a list
    overlaps = list(set(overlaps))
    # Determine number of collisions
    n_overlaps = len(overlaps)
    """ Now checking for overlappings with walls """
    # Wall in positive x direction
    if (x_i + d_i / 2) >= config.A_0:
        walls = 1
    # Wall in negative x direction
    elif (x_i - d_i / 2) <= -config.A_0:
        walls = -1
    # Wall in positive y direction
    elif (y_i + d_i / 2) >= config.A_0:
        walls = 2
    # Wall in negative y direction
    elif (y_i - d_i / 2) <= -config.A_0:
        walls = -2
    # Wall in positive x direction and positive y direction    
    elif (x_i + d_i / 2) >= config.A_0 and (y_i + d_i / 2) >= config.A_0:
        walls = 3
    # Wall in negative x direction and positive y direction 
    elif (x_i - d_i / 2) <= -config.A_0 and (y_i + d_i / 2) >= config.A_0:
        walls = -3
    # Wall in positive x direction and negative y direction 
    elif (x_i + d_i / 2) >= config.A_0 and (y_i - d_i / 2) <= -config.A_0:
        walls = 4
    # Wall in negative x direction and negaive y direction 
    elif (x_i - d_i / 2) <= -config.A_0 and (y_i - d_i / 2) <= -config.A_0:
        walls = -4

    return n_overlaps, overlaps, walls

def position_update1(y_i, x_i, d_i, M_particle_steady, overlaps):
    """ Function that updates the position of the considered particle overlapping with one steady state particle. """
    # Coordinates of the steady state particle 
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the steady state particle 
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Required distance between the centers of the considered particle and the steady state particle
    dist_1i = (d1 + d_i) / 2  
    # Actual distance between the centers of the considered particle and the steady state particle in x and y coordinate
    delta_x = x_i - x1
    delta_y = y_i - y1
    dist_1i_current = np.sqrt(delta_x**2 + delta_y**2)
    # Calculate the scaling factor
    scaling_factor = dist_1i / dist_1i_current
    # Calculate the new position
    x = x1 + delta_x * scaling_factor
    y = y1 + delta_y * scaling_factor

    return y, x

def position_update2(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps):
    """
    Function to update the position of the considered particle if it overlaps with two particles.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, an intersection circle plane is calculated corresponding to the intersection area of two auxiliary spheres.
    The intersection plane is used to identify the guesses of the both solutions of the considered particle's new position.
    Moreover, the plane is used to check wheter the solver solutions are correct.
    The centers of the auxiliary spheres corresponds to the center coordinates of the both steady state particles.
    The radii of the auxiliary spheres corresponds to the sums of the steady state particles' radii and the considered particle's radius.
    To identify the logical solution of the both solutions, the distances between the original position and the new positions are calculated.
    The new position with the lowest distance to the original position is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update2 = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the first particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Center coordinates of the first auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Radius of the first auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Diameter of the second particle in steady state
    d2 = float(M_particle_steady[overlaps[1], 5])
    # Center coordinates of the second auxiliary sphere
    c2 = np.array([z2, y2, x2])
    # Radius of the second auxiliary sphere 
    r2 = (d_i + d2) / 2
    """ Calculation of the intersection circle plane """  
    # Vector between the centers
    V_centers_12 = c2 - c1
    # Distance between the centers
    dist_centers_12 = np.linalg.norm(V_centers_12)
    # Normal vector of the intersection circle plane
    V_unit_centers_12 = V_centers_12 / dist_centers_12
    # Distance from the first sphere's center to the intersection circle plane
    dist_c1_plane = (r1**2 - r2**2 + dist_centers_12**2) / (2 * dist_centers_12)
    # Center of the intersection circle (point on the intersection plane closest to sphere1)
    c_ic = c1 + dist_c1_plane * V_unit_centers_12
    """ 
    Calculation of the reflected position of the considered particle.
    First, the vector between the centers of the considered particle and the intersection circle is determined.
    Then, this vector is reflected such a way that the projection of this vector on the yx plane is rotated by 180 degrees on the yx plane.
    The coordiantes of the reflected position are used as second guess for the solver to determine the second solver solution. 
    """
    # Vector between the centers of the considered particle and the intersection circle
    V_centers_ici = np.array([z_i, y_i, x_i]) - np.array(c_ic)
    # Reflected vector
    V_reflected = np.array([V_centers_ici[0], -V_centers_ici[1], -V_centers_ici[2]])
    # Reflected position
    P_reflected = np.array(c_ic) + V_reflected
    # Coordinates for the initial guess
    x_01 = x_i
    y_01 = y_i
    x_02 = P_reflected[2]
    y_02 = P_reflected[1]
    # Initial guess values
    initial_guess1 = [x_01, y_01]
    initial_guess2 = [x_02, y_02]
    """ 
    Solver to find possible new positions of the considered particle
    The first guess is the logical one. The second is just to determine the second solver solution.
    The second solver solution is just to check wheter the first one is the right one 
    """
    x_s1, y_s1 = fsolve(solver_position2, initial_guess1, args=(c1, c2, r1, r2, z_i))
    x_s2, y_s2 = fsolve(solver_position2, initial_guess2, args=(c1, c2, r1, r2, z_i))
    """ Check wheter the solver found two different solutions or the same twice """
    # Error value is 1 if the solver found the same position twice
    if x_s1 == x_s2 and y_s1 == y_s2:
        error_position_update2 = 1
    # Exclude values behind the simulation walls
    if x_s1 < -config.A_0 or x_s1 > config.A_0:
        x_s1 = None
    if x_s2 < -config.A_0 or x_s2 > config.A_0:
        x_s2 = None
    if y_s1 < -config.A_0 or y_s1 > config.A_0:
        y_s1 = None
    if y_s2 < -config.A_0 or y_s2 > config.A_0:
        y_s2 = None    
    """ 
    Determine the actual new position of the considered particle from the both solver solutions.
    Must be the closest position to the original position of the considered particle
    """
    # Choose the new position based on the validity of values
    if y_s1 == None or x_s1 == None:
        if y_s2 != None and x_s2 != None:
            x = x_s2
            y = y_s2
        else:
            error_position_update2 = 3
            return y_i, x_i, error_position_update2   
    elif y_s2 == None or x_s2 == None:
        if y_s1 != None or x_s1 != None:
            x = x_s1
            y = y_s1
        else: 
            error_position_update2 = 3
            return y_i, x_i, error_position_update2
    else:
        # Vectors between the centers of the considered particle's original position and the new positions
        V_centers_new1 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s1, x_s1])
        V_centers_new2 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s2, x_s2])
        # Distances between these centers
        dist_centers_new1 = np.linalg.norm(V_centers_new1)
        dist_centers_new2 = np.linalg.norm(V_centers_new2)
        # Choose new position based on the shortest distance
        if dist_centers_new1 <= dist_centers_new2:
            x = x_s1
            y = y_s1
        else:
            x = x_s2
            y = y_s2
    """ 
    Check wheter the new positions' coordinates are located on the intersection plane 
    Just double checking wheter the solver found correct solutions
    The equation of the intersection plane is Az+By+Cx+D=0, where A, B, C and D are the equation coefficients.
    """
    # Determine the plane equation coefficients
    # The normal vector to the plane is the same as the V_centers_12
    coeff_A, coeff_B, coeff_C = V_unit_centers_12
    # Calculate D using the point on the plane
    coeff_D = -(coeff_A * c_ic[0] + coeff_B * c_ic[1] + coeff_C * c_ic[2])
    # Check whether the new positions of considered particle is located on the intersection plane
    on_plane = check_point_on_plane(z_i, y, x, coeff_A, coeff_B, coeff_C, coeff_D)
    # Error value is 2 if the position of the first solver solution is not located on the intersection plane
    if on_plane == False:
        error_position_update2 = 2

    return y, x, error_position_update2

def solver_position2(fit_pars, c1, c2, r1, r2, z_i):
    """ Solver to calculate the new position of the considered particle if it overlaps with two particles """
    # Solver variables
    x, y = fit_pars
    # Equation of a circle. The circles coresponds here to the auxiliary spheres
    eq1 = (x - c1[2])**2 + (y - c1[1])**2 + (z_i - c1[0])**2 - r1**2
    eq2 = (x - c2[2])**2 + (y - c2[1])**2 + (z_i - c2[0])**2 - r2**2
    return [eq1, eq2]

def check_point_on_plane(z_i, y, x, coeff_A, coeff_B, coeff_C, coeff_D): 
    """ Function that checks whether a point lies on a plane or not (plane equation in coordinate form) """
    # Calculate the left-hand side of the plane equation
    check_zero = z_i*coeff_A + y*coeff_B + x*coeff_C + coeff_D
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    on_plane = abs(check_zero) < 1e-9
    return on_plane

def position_update3(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps):
    """
    Function to update the position of the considered particle if it overlaps with three particles.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, a triangle plane is calculated which crosses the centers of the three steady state particles.
    Moreover, the centroid of the triangle is determined.
    The triangle plane is used to identify the guesses of the both solutions of the considered particle's new position.
    It is tested whether the both solutions differ and if they are correct.
    To identify the logical solution of the both solutions, the z-coordinates of the both soultions and the centroid are compared.
    The new position with the z-coordinate above the centroid is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update3 = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the first particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Center coordinates of the first auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Radius of the first auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Diameter of the second particle in steady state
    d2 = float(M_particle_steady[overlaps[1], 5])
    # Center coordinates of the second auxiliary sphere
    c2 = np.array([z2, y2, x2])
    # Radius of the second auxiliary sphere 
    r2 = (d_i + d2) / 2
    # Coordinates of the third particle in steady state
    z3 = float(M_particle_steady[overlaps[2], 1])
    y3 = float(M_particle_steady[overlaps[2], 2])
    x3 = float(M_particle_steady[overlaps[2], 3])
    # Diameter of the third particle in steady state
    d3 = float(M_particle_steady[overlaps[2], 5])
    # Center coordinates of the third auxiliary sphere
    c3 = np.array([z3, y3, x3])
    # Radius of the third auxiliary sphere 
    r3 = (d_i + d3) / 2
    """ Calculation of the triangle's plane """
    # Calculate vectors AB and AC
    AB = np.array([z2 - z1, y2 - y1, x2 - x1])
    AC = np.array([z3 - z1, y3 - y1, x3 - x1])
    # Compute the normal vector to the plane (cross product of AB and AC)
    V_triangle_norm = np.cross(AB, AC)
    A, B, C = V_triangle_norm
    # Calculate D using the plane equation Ax + By + Cz + D = 0
    # Using point A (x1, y1, z1) to find D
    D = -(A * z1 + B * y1 + C * x1)
    """ 
    Calculation of the reflected position of the considered particle.
    First, the distance between the center of the considered particle and the triangle's plane is determined.
    Then, the position of the considered particle is reflected about the triangle's plane
    The coordiantes of the reflected position are used as second guess for the solver to determine the second solver solution. 
    """
    # Calculate the distance from the considered particle's center to the plane
    distance_pi = (A * z_i + B * y_i + C * x_i + D) / np.linalg.norm(V_triangle_norm)
    # Calculate the reflection point on the plane
    P_reflected = np.array([z_i, y_i, x_i]) - 2 * distance_pi * V_triangle_norm / np.linalg.norm(V_triangle_norm)
    """ Solver to find possible new position of the considered particle """
    # Coordinates for the initial guess
    x_01 = x_i
    y_01 = y_i
    z_01 = z_i
    x_02 = P_reflected[2]
    y_02 = P_reflected[1]
    z_02 = P_reflected[0]
    # Initial guess values
    initial_guess1 = [x_01, y_01, z_01]
    initial_guess2 = [x_02, y_02, z_02]
    # Solver
    x_s1, y_s1, z_s1 = fsolve(solver_position3, initial_guess1, args=(c1, c2, c3, r1, r2, r3))
    x_s2, y_s2, z_s2 = fsolve(solver_position3, initial_guess2, args=(c1, c2, c3, r1, r2, r3))
    """ Check wheter the solver found two different solutions or the same twice """
    # Error value is 1 if the solver found the same position twice
    if x_s1 == x_s2 and y_s1 == y_s2 and z_s1 == z_s2:
        error_position_update3 = 1
    # Exclude values behind the simulation walls
    if x_s1 < -config.A_0 or x_s1 > config.A_0:
        x_s1 = None
    if x_s2 < -config.A_0 or x_s2 > config.A_0:
        x_s2 = None
    if y_s1 < -config.A_0 or y_s1 > config.A_0:
        y_s1 = None
    if y_s2 < -config.A_0 or y_s2 > config.A_0:
        y_s2 = None 
    """ 
    Determine the actual new position of the considered particle from the both solver solutions.
    Must be the closest position to the original position of the considered particle
    """
    # Z-coordinate of polygon's centroid    
    z_cen = (z1 + z2 + z3) / 3
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the first solver solution are assigned to the center of the considered particle
    if x_s1 != None and y_s1 != None and z_s1 > z_cen:
        z = z_s1
        y = y_s1
        x = x_s1
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the second solver solution are assigned to the center of the considered particle
    elif x_s2 != None and y_s2 != None and z_s2 > z_cen: 
        z = z_s2
        y = y_s2
        x = x_s2
    else: 
        error_position_update3 = 2
        return z_i, y_i, x_i, error_position_update3
    """ 
    Check wheter the distances between the centers of the three steady state particles 
    and the new center position of the considered particle are in the samerange as the sums of their radii.
    """
    same_range = check_distances3(r1, r2, r3, c1, c2, c3, z, y, x)
    # Error value is 2 if the distance between the steady state particles and the position of the solver solution 
    # is not in the same range as the sums of their radii
    if same_range == False:
        error_position_update3 = 3
        return z_i, y_i, x_i, error_position_update3

    return z, y, x, error_position_update3

def solver_position3(fit_pars, c1, c2, c3, r1, r2, r3):
    """ Solver to calculate the new position of the considered particle if it overlaps with three particles """
    # Solver variables
    x, y, z = fit_pars
    # Equation of a circle. The circles coresponds here to the auxiliary spheres
    eq1 = (x - c1[2])**2 + (y - c1[1])**2 + (z - c1[0])**2 - r1**2
    eq2 = (x - c2[2])**2 + (y - c2[1])**2 + (z - c2[0])**2 - r2**2
    eq3 = (x - c3[2])**2 + (y - c3[1])**2 + (z - c3[0])**2 - r3**2
    return [eq1, eq2, eq3]

def check_distances3(r1, r2, r3, c1, c2, c3, z_s, y_s, x_s): 
    """ Function that checks wheter the distances between the centers of the three steady state particles 
    and the new center position of the considered particle are in the same range as the sums of their radii. """
    # True if distances are in the same range as the sums of radii
    same_range = True
    # Vectors between the centers of the steady state particles and the new center positions of the considered particle    
    V_centers_1i = np.array([z_s, y_s, x_s]) - c1
    V_centers_2i = np.array([z_s, y_s, x_s]) - c2
    V_centers_3i = np.array([z_s, y_s, x_s]) - c3
    # Distance between the centers of the steady state particles and the new center position of the considered particle
    dist_centers_1i = np.linalg.norm(V_centers_1i)
    dist_centers_2i = np.linalg.norm(V_centers_2i)
    dist_centers_3i = np.linalg.norm(V_centers_3i)
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    if abs(dist_centers_1i-r1) > 1e-9 or abs(dist_centers_2i-r2) > 1e-9 or abs(dist_centers_3i-r3) > 1e-9:
        same_range = False

    return same_range

def check_steadystate3(z_i, y_i, x_i, M_particle_steady, overlaps):
    """ 
    Function to check the steady state of the considered particle if it is in contact with three steady state particles.
    As soon as considered particle collides with three steady state particles, a steady state check is performed.
    Considered particle is in steady state if its center lies within the y-x-plane of a triangle located between the 
    steady state particle centers. 
    Barycentric coordinate method is used to determine if the center of the considered particle lies within the polygon.
    Moreover, barycentric coordinates are used to determine the subdomain of the triangle where the center lies.
    Based on the subdomain, the particle rolling direction is identified.
    """
    # Used in case an error occurs during steadystate check.
    error_check_steadystate3 = 0
    # Variable for steady state check condition
    steady_check = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Coordinates of the third particle in steady state
    z3 = float(M_particle_steady[overlaps[2], 1])
    y3 = float(M_particle_steady[overlaps[2], 2])
    x3 = float(M_particle_steady[overlaps[2], 3])
    # Z-coordinate of triangle's centroid
    z_cen = (z1 + z2 + z3) / 3
    # Check whether the center of the considered particle is located above the centroid
    # Double checking at this point, since already handled in the position_update3 function.
    # Error value is 1 when the center is located below the centroid
    if z_i < z_cen:
        error_check_steadystate3 = 1
        return steady_check, overlaps, error_check_steadystate3
    """ Barycentric coordinate technique """
    # Denominator
    denominator = ((y2 - y3) * (x1 - x3) + (x3 - x2) * (y1 - y3))
    # Check whether the triangle points are collinear
    # Error value is 2 when the triangle points are collinear
    if denominator == 0:
        error_check_steadystate3 = 2  
        return steady_check, overlaps, error_check_steadystate3
    # Barycentric coordinates
    a = ((y2 - y3) * (x_i - x3) + (x3 - x2) * (y_i - y3)) / denominator
    b = ((y3 - y1) * (x_i - x3) + (x1 - x3) * (y_i - y3)) / denominator
    c = 1 - a - b
    # Determine the subdomain of the triangle and thus wheter it is in steady state or in which direction it is rolling
    if (0 <= a <= 1) and (0 <= b <= 1) and (0 <= c <= 1):
        steady_check = 3
    elif a < 0 and b >= 0 and c >= 0:
        # Rolling along steady state particle Nr.2 and Nr.3
        overlaps = np.delete(overlaps, 0)
        steady_check = 2
    elif b < 0 and a >= 0 and c >= 0:
        # Rolling along steady state particle Nr.1 and Nr.3
        overlaps = np.delete(overlaps, 1)
        steady_check = 2
    elif c < 0 and a >= 0 and b >= 0:
        # Rolling along steady state particle Nr.1 and Nr.2
        overlaps = np.delete(overlaps, 2)
        steady_check = 2
    elif a < 0 and b < 0 and c > 1:
        # Rolling along steady state particle Nr.3
        overlaps = np.delete(overlaps, [0, 1])
        steady_check = 1
    elif a < 0 and c < 0 and b > 1:
        # Rolling along steady state particle Nr.2
        overlaps = np.delete(overlaps, [0, 2])
        steady_check = 1
    elif b < 0 and c < 0 and a > 1:
        # Rolling along steady state particle Nr.1 
        overlaps = np.delete(overlaps, [1, 2])
        steady_check = 1
    else:
        # Error value is 3 when the subdomain of the triangle could not be determined correctly
        error_check_steadystate3 = 3

    return steady_check, overlaps, error_check_steadystate3

def position_update1w(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps, walls):
    """
    Function to update the position of the considered particle if it overlaps with one particle and one wall.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, an intersection circle plane is calculated corresponding to the intersection area of one auxiliary sphere and the wall.
    The intersection plane is used to identify the guesses of the both solutions of the considered particle's new position.
    Moreover, the plane is used to check wheter the solver solutions are correct.
    The center of the auxiliary sphere corresponds to the center coordinates of the steady state particle.
    The radius of the auxiliary sphere corresponds to the sums of the steady state particle's radius and the considered particle's radius.
    To identify the logical solution of the both solutions, the distances between the original position and the new positions are calculated.
    The new position with the lowest distance to the original position is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update1w = 0
    # Coordinates of the particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Center coordinates of the auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Radius of the auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Coordinates of the wall
    z_w1 = z1
    if walls == 1:
        x_w1 = config.A_0
        y_w1 = y1
    elif walls == -1:
        x_w1 = -config.A_0
        y_w1 = y1
    elif walls == 2:
        x_w1 = x1
        y_w1 = config.A_0
    elif walls == -2:
        x_w1 = x1
        y_w1 = -config.A_0
    # Center coordinates of the wall
    c_w1 = np.array([z_w1, y_w1, x_w1])
    # Distance between wall and center of considered particle if they are in contact
    dist_wi = d_i / 2
    """ Calculation of the intersection circle plane """  
    # Vector between the centers
    V_centers_1w = c_w1 - c1
    # Distance between the centers
    dist_centers_1w = np.linalg.norm(V_centers_1w)
    # Normal vector of the intersection circle plane
    V_unit_centers_1w = V_centers_1w / dist_centers_1w
    # Distance from the first sphere's center to the intersection circle plane
    dist_c1_plane = dist_centers_1w - dist_wi
    # Center of the intersection circle (point on the intersection plane closest to sphere1)
    c_ic = c1 + dist_c1_plane * V_unit_centers_1w
    """ 
    Calculation of the reflected position of the considered particle.
    First, the vector between the centers of the considered particle and the intersection circle is determined.
    Then, this vector is reflected such a way that the projection of this vector on the yx plane is rotated by 180 degrees on the yx plane.
    The coordiantes of the reflected position are used as second guess for the solver to determine the second solver solution. 
    """
    # Vector between the centers of the considered particle and the intersection circle
    V_centers_ici = np.array([z_i, y_i, x_i]) - np.array(c_ic)
    # Reflected vector
    V_reflected = np.array([V_centers_ici[0], -V_centers_ici[1], -V_centers_ici[2]])
    # Reflected position
    P_reflected = np.array(c_ic) + V_reflected
    # Coordinates for the initial guess
    x_01 = x_i
    y_01 = y_i
    x_02 = P_reflected[2]
    y_02 = P_reflected[1]
    # If contact is with X-wall
    if walls == 1 or walls == -1:
        # Initial guess values
        initial_guess1 = y_01
        initial_guess2 = y_02
        # X-coordinate of the new position
        if walls == 1:
            x_s = config.A_0 - d_i / 2
        elif walls == -1:
            x_s = -config.A_0 + d_i / 2   
        """ 
        Solver to find possible new positions of the considered particle
        The first guess is the logical one. The second is just to determine the second solver solution.
        The second solver solution is just to check wheter the first one is the right one 
        """
        y_s1 = fsolve(solver_position1w, initial_guess1, args=(c1, r1, z_i, x_s, walls))
        y_s2 = fsolve(solver_position1w, initial_guess2, args=(c1, r1, z_i, x_s, walls))
        y_s1 = y_s1[0]
        y_s2 = y_s2[0]
        """ Check wheter the solver found two different solutions or the same twice """
        # Error value is 1 if the solver found the same position twice
        if y_s1 == y_s2:
            error_position_update1w = 1
        # Exclude values behind the simulation walls
        if y_s1 < -config.A_0 or y_s1 > config.A_0:
            y_s1 = None
        if y_s2 < -config.A_0 or y_s2 > config.A_0:
            y_s2 = None
        """ 
        Determine the actual new position of the considered particle from the both solver solutions.
        Must be the closest position to the original position of the considered particle
        """
        # Coose the new position based on the validity of values
        if y_s1 == None and y_s2 != None:
            y = y_s2
        elif y_s2 == None and y_s1 != None:
            y = y_s1
        elif y_s1 == None and y_s2 == None:
            error_position_update1w = 3
            return y_i, x_i, error_position_update1w
        else:
            # Vectors between the centers of the considered particle's original position and the new positions
            V_centers_new1 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s1, x_s])
            V_centers_new2 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s2, x_s])
            # Distances between these centers
            dist_centers_new1 = np.linalg.norm(V_centers_new1)
            dist_centers_new2 = np.linalg.norm(V_centers_new2)
            # Choose new position based on the shortest distance
            if dist_centers_new1 <= dist_centers_new2:
                y = y_s1
            else:
                y = y_s2
        x = x_s
    # If contact is with Y-wall
    elif walls == 2 or walls == -2:
        # Initial guess values
        initial_guess1 = x_01
        initial_guess2 = x_02
        # Y-coordinate of the new position
        if walls == 2:
            y_s = config.A_0 - d_i / 2
        elif walls == -2:
            y_s = -config.A_0 + d_i / 2   
        """ 
        Solver to find possible new positions of the considered particle
        The first guess is the logical one. The second is just to determine the second solver solution.
        The second solver solution is just to check wheter the first one is the right one 
        """
        x_s1 = fsolve(solver_position1w, initial_guess1, args=(c1, r1, z_i, y_s, walls))
        x_s2 = fsolve(solver_position1w, initial_guess2, args=(c1, r1, z_i, y_s, walls))
        x_s1 = x_s1[0]
        x_s2 = x_s2[0]
        """ Check wheter the solver found two different solutions or the same twice """
        # Error value is 1 if the solver found the same position twice
        if x_s1 == x_s2:
            error_position_update1w = 1
        # Exclude values behind the simulation walls
        if x_s1 < -config.A_0 or x_s1 > config.A_0:
            x_s1 = None
        if x_s2 < -config.A_0 or x_s2 > config.A_0:
            x_s2 = None
        """ 
        Determine the actual new position of the considered particle from the both solver solutions.
        Must be the closest position to the original position of the considered particle
        """
        # Coose the new position based on the validity of values
        if x_s1 == None and x_s2 != None:
            x = x_s2
        elif x_s2 == None and x_s1 != None:
            x = x_s1
        elif x_s1 == None and x_s2 == None:
            error_position_update1w = 3
            return y_i, x_i, error_position_update1w
        else:   
            # Vectors between the centers of the considered particle's original position and the new positions
            V_centers_new1 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s, x_s1])
            V_centers_new2 = np.array([z_i, y_i, x_i]) - np.array([z_i, y_s, x_s2])
            # Distances between these centers
            dist_centers_new1 = np.linalg.norm(V_centers_new1)
            dist_centers_new2 = np.linalg.norm(V_centers_new2)
            # Choose new position based on the shortest distance
            if dist_centers_new1 <= dist_centers_new2:
                x = x_s1
            else:
                x = x_s2
        # Assign x-coordinate
        y = y_s
    """ 
    Check wheter the new positions' coordinates are located on the intersection plane 
    Just double checking wheter the solver found correct solutions
    The equation of the intersection plane is Az+By+Cx+D=0, where A, B, C and D are the equation coefficients.
    """
    # Determine the plane equation coefficients
    # The normal vector to the plane is the same as the V_centers_12
    coeff_A, coeff_B, coeff_C = V_unit_centers_1w
    # Calculate D using the point on the plane
    coeff_D = -(coeff_A * c_ic[0] + coeff_B * c_ic[1] + coeff_C * c_ic[2])
    # Check whether the new positions of considered particle is located on the intersection plane
    on_plane = check_point_on_plane(z_i, y, x, coeff_A, coeff_B, coeff_C, coeff_D)
    # Error value is 2 if the position of the first solver solution is not located on the intersection plane
    if on_plane == False:
        error_position_update1w = 2
        return y_i, x_i, error_position_update1w

    return y, x, error_position_update1w

def solver_position1w(fit_pars, c1, r1, z_i, par, walls):
    """ Solver to calculate the new position of the considered particle if it overlaps with one particle and one wall """
    if walls == 1 or walls == -1:
        # Solver variables
        y = fit_pars
        # Equation of a circle. The circles coresponds here to the auxiliary spheres
        eq1 = (par - c1[2])**2 + (y - c1[1])**2 + (z_i - c1[0])**2 - r1**2
        return eq1
    elif walls == 2 or walls == -2:
        # Solver variables
        x = fit_pars
        # Equation of a circle. The circles coresponds here to the auxiliary spheres
        eq1 = (x - c1[2])**2 + (par - c1[1])**2 + (z_i - c1[0])**2 - r1**2
        return eq1  

def position_update2w(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps, walls):
    """
    Function to update the position of the considered particle if it overlaps with two particles and one wall.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, a triangle plane is calculated which crosses the centers of the two steady state particles and the wall.
    Moreover, the centroid of the triangle is determined.
    The triangle plane is used to identify the guesses of the both solutions of the considered particle's new position.
    It is tested whether the both solutions differ and if they are correct.
    To identify the logical solution of the both solutions, the z-coordinates of the both soultions and the centroid are compared.
    The new position with the z-coordinate above the centroid is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update2w = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the first particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Center coordinates of the first auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Radius of the first auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Diameter of the second particle in steady state
    d2 = float(M_particle_steady[overlaps[1], 5])
    # Center coordinates of the second auxiliary sphere
    c2 = np.array([z2, y2, x2])
    # Radius of the second auxiliary sphere 
    r2 = (d_i + d2) / 2
    # Coordinates of the wall
    z_w1 = (z1+z2)/2
    if walls == 1:
        x_w1 = config.A_0
        y_w1 = (y1+y2)/2
    elif walls == -1:
        x_w1 = -config.A_0
        y_w1 = (y1+y2)/2
    elif walls == 2:
        x_w1 = (x1+x2)/2
        y_w1 = config.A_0
    elif walls == -2:
        x_w1 = (x1+x2)/2
        y_w1 = -config.A_0
    """ Calculation of the triangle's plane """
    # Calculate vectors AB and AC
    AB = np.array([z2 - z1, y2 - y1, x2 - x1])
    AC = np.array([z_w1 - z1, y_w1 - y1, x_w1 - x1])
    # Compute the normal vector to the plane (cross product of AB and AC)
    V_triangle_norm = np.cross(AB, AC)
    A, B, C = V_triangle_norm
    # Calculate D using the plane equation Ax + By + Cz + D = 0
    # Using point A (x1, y1, z1) to find D
    D = -(A * z1 + B * y1 + C * x1)
    """ 
    Calculation of the reflected position of the considered particle.
    First, the distance between the center of the considered particle and the triangle's plane is determined.
    Then, the position of the considered particle is reflected about the triangle's plane
    The coordiantes of the reflected position are used as second guess for the solver to determine the second solver solution. 
    """
    # Calculate the distance from the considered particle's center to the plane
    distance_pi = (A * z_i + B * y_i + C * x_i + D) / np.linalg.norm(V_triangle_norm)
    # Calculate the reflection point on the plane
    P_reflected = np.array([z_i, y_i, x_i]) - 2 * distance_pi * V_triangle_norm / np.linalg.norm(V_triangle_norm)
    """ Solver to find possible new position of the considered particle """
    # Coordinates for the initial guess
    x_01 = x_i
    y_01 = y_i
    z_01 = z_i
    x_02 = P_reflected[2]
    y_02 = P_reflected[1]
    z_02 = P_reflected[0]
    # If contact is with X-wall
    if walls == 1 or walls == -1:
        # Initial guess values
        initial_guess1 = [y_01, z_01]
        initial_guess2 = [y_02, z_02]
        # X-coordinate of the new position
        if walls == 1:
            x_s1 = config.A_0 - d_i / 2
            x_s2 = config.A_0 - d_i / 2
        elif walls == -1:
            x_s1 = -config.A_0 + d_i / 2 
            x_s2 = -config.A_0 + d_i / 2
        # Solver
        y_s1, z_s1 = fsolve(solver_position2w, initial_guess1, args=(c1, c2, r1, r2, x_s1, walls))
        y_s2, z_s2 = fsolve(solver_position2w, initial_guess2, args=(c1, c2, r1, r2, x_s2, walls))
        """ Check wheter the solver found two different solutions or the same twice """
        # Error value is 1 if the solver found the same position twice
        if y_s1 == y_s2 and z_s1 == z_s2:
            error_position_update2w = 1
        # Exclude values behind the simulation walls
        if y_s1 < -config.A_0 or y_s1 > config.A_0:
            y_s1 = None
        if y_s2 < -config.A_0 or y_s2 > config.A_0:
            y_s2 = None    
    # If contact is with Y-wall
    if walls == 2 or walls == -2:    
        # Initial guess values
        initial_guess1 = [x_01, z_01]
        initial_guess2 = [x_02, z_02]
        # X-coordinate of the new position
        if walls == 2:
            y_s1 = config.A_0 - d_i / 2
            y_s2 = config.A_0 - d_i / 2
        elif walls == -2:
            y_s1 = -config.A_0 + d_i / 2 
            y_s2 = -config.A_0 + d_i / 2
        # Solver
        x_s1, z_s1 = fsolve(solver_position2w, initial_guess1, args=(c1, c2, r1, r2, y_s1, walls))
        x_s2, z_s2 = fsolve(solver_position2w, initial_guess2, args=(c1, c2, r1, r2, y_s2, walls))
        """ Check wheter the solver found two different solutions or the same twice """
        # Error value is 1 if the solver found the same position twice
        if x_s1 == x_s2 and z_s1 == z_s2:
            error_position_update2w = 1       
        # Exclude values behind the simulation walls
        if x_s1 < -config.A_0 or x_s1 > config.A_0:
            y_s1 = None
        if x_s2 < -config.A_0 or x_s2 > config.A_0:
            x_s2 = None
    """ 
    Determine the actual new position of the considered particle from the both solver solutions.
    Must be the closest position to the original position of the considered particle
    """
    # Z-coordinate of triangles's centroid    
    z_cen = (z1 + z2 + z_w1) / 3
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the first solver solution are assigned to the center of the considered particle
    if x_s1 != None and y_s1 != None and z_s1 > z_cen:
        z = z_s1
        y = y_s1
        x = x_s1
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the second solver solution are assigned to the center of the considered particle
    elif x_s2 != None and y_s2 != None and z_s2 > z_cen: 
        z = z_s2
        y = y_s2
        x = x_s2
    else: 
        error_position_update2w = 2
        return z_i, y_i, x_i, error_position_update2w
    """ 
    Check wheter the distances between the centers of the three steady state particles 
    and the new center position of the considered particle are in the samerange as the sums of their radii.
    """
    same_range = check_distances2w(config, r1, r2, c1, c2, d_i, z, y, x, walls)
    # Error value is 2 if the distance between the steady state particles and the position of the solver solution 
    # is not in the same range as the sums of their radii
    if same_range == False:
        error_position_update2w = 3
        return z_i, y_i, x_i, error_position_update2w

    return z, y, x, error_position_update2w

def solver_position2w(fit_pars, c1, c2, r1, r2, par, walls):
    """ Solver to calculate the new position of the considered particle if it overlaps with two particles and one wall """
    if walls == 1 or walls == -1:
        # Solver variables
        y, z = fit_pars
        # Equation of a circle. The circles coresponds here to the auxiliary spheres
        eq1 = (par - c1[2])**2 + (y - c1[1])**2 + (z - c1[0])**2 - r1**2
        eq2 = (par - c2[2])**2 + (y - c2[1])**2 + (z - c2[0])**2 - r2**2
        return [eq1, eq2]
    elif walls == 2 or walls == -2:
        # Solver variables
        x, z = fit_pars
        # Equation of a circle. The circles coresponds here to the auxiliary spheres
        eq1 = (x - c1[2])**2 + (par - c1[1])**2 + (z - c1[0])**2 - r1**2
        eq2 = (x - c2[2])**2 + (par - c2[1])**2 + (z - c2[0])**2 - r2**2
        return [eq1, eq2]

def check_distances2w(config, r1, r2, c1, c2, d_i, z_s, y_s, x_s, walls): 
    """ Function that checks wheter the distances between the centers of the two steady state particles, the wall, 
    and the new center position of the considered particle are in the same range as the sums of their radii. """
    # True if distances are in the same range as the sums of radii
    same_range = True
    # Vectors between the centers of the steady state particles and the new center positions of the considered particle    
    V_centers_1i = np.array([z_s, y_s, x_s]) - c1
    V_centers_2i = np.array([z_s, y_s, x_s]) - c2
    # Distance between the centers of the steady state particles and the new center position of the considered particle
    dist_centers_1i = np.linalg.norm(V_centers_1i)
    dist_centers_2i = np.linalg.norm(V_centers_2i)
    # Distance between the center of the considered particle and the wall
    if walls == 1 or walls == -1: 
        dist_wi = config.A_0 - abs(x_s) 
    elif walls == 2 or walls == -2:
        dist_wi = config.A_0 - abs(y_s)
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    if abs(dist_centers_1i-r1) > 1e-9 or abs(dist_centers_2i-r2) > 1e-9 or abs(dist_wi-d_i/2) > 1e-9:
        same_range = False

    return same_range

def check_steadystate2w(config, z_i, y_i, x_i, M_particle_steady, overlaps, walls):
    """ 
    Function to check the steady state of the considered particle if it is in contact with two steady state particles and one wall.
    As soon as considered particle collides with two steady state particles and one wall, a steady state check is performed.
    Considered particle is in steady state if its center lies within the y-x-plane of a triangle located between the 
    steady state particle centers and the wall. 
    Barycentric coordinate method is used to determine if the center of the considered particle lies within the polygon.
    Moreover, barycentric coordinates are used to determine the subdomain of the triangle where the center lies.
    Based on the subdomain, the particle rolling direction is identified.
    """
    # Used in case an error occurs during steadystate check.
    error_check_steadystate3 = 0
    # Variable for steady state check condition
    steady_check = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Coordinates of the wall
    z_w1 = z_i
    if walls == 1:
        x_w1 = config.A_0
        y_w1 = y_i
    elif walls == -1:
        x_w1 = -config.A_0
        y_w1 = y_i
    elif walls == 2:
        x_w1 = x_i
        y_w1 = config.A_0
    elif walls == -2:
        x_w1 = x_i
        y_w1 = -config.A_0
    # Z-coordinate of triangle's centroid
    z_cen = (z1 + z2 + z_w1) / 3
    # Check whether the center of the considered particle is located above the centroid
    # Double checking at this point, since already handled in the position_update3 function.
    # Error value is 1 when the center is located below the centroid
    if z_i < z_cen:
        error_check_steadystate3 = 1
        return steady_check, overlaps, error_check_steadystate3
    """ Barycentric coordinate technique """
    # Denominator
    denominator = ((y2 - y_w1) * (x1 - x_w1) + (x_w1 - x2) * (y1 - y_w1))
    # Check whether the triangle points are collinear
    # Error value is 2 when the triangle points are collinear
    if denominator == 0:
        error_check_steadystate3 = 2  
        return steady_check, overlaps, error_check_steadystate3
    # Barycentric coordinates
    a = ((y2 - y_w1) * (x_i - x_w1) + (x_w1 - x2) * (y_i - y_w1)) / denominator
    b = ((y_w1 - y1) * (x_i - x_w1) + (x1 - x_w1) * (y_i - y_w1)) / denominator
    c = 1 - a - b
    # Determine the subdomain of the triangle and thus wheter it is in steady state or in which direction it is rolling
    if (0 <= a <= 1) and (0 <= b <= 1) and (0 <= c <= 1):
        steady_check = 3
    elif a < 0 and b >= 0 and c >= 0:
        # Rolling along steady state particle Nr.2 and wall
        overlaps = np.delete(overlaps, 0)
        steady_check = 4
    elif b < 0 and a >= 0 and c >= 0:
        # Rolling along steady state particle Nr.1 and wall
        overlaps = np.delete(overlaps, 1)
        steady_check = 4
    elif c < 0 and a >= 0 and b >= 0:
        # Rolling along steady state particle Nr.1 and Nr.2
        steady_check = 2
    elif a < 0 and c < 0 and b > 1:
        # Rolling along steady state particle Nr.2 
        overlaps = np.delete(overlaps, 0)
        steady_check = 1
    elif b < 0 and c < 0 and a > 1:
        # Rolling along steady state particle Nr.1 
        overlaps = np.delete(overlaps, 1)
        steady_check = 1
    else:
        # Error value is 3 when the subdomain of the triangle could not be determined correctly
        error_check_steadystate3 = 3

    return steady_check, overlaps, error_check_steadystate3

def position_update1w2(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps, walls):
    """
    Function to update the position of the considered particle if it overlaps with one particle and two walls.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, a triangle plane is calculated which crosses the center of the steady state particle and the walls.
    Moreover, the centroid of the triangle is determined.
    The triangle plane is used to identify the guesses of the both solutions of the considered particle's new position.
    It is tested whether the both solutions differ and if they are correct.
    To identify the logical solution of the both solutions, the z-coordinates of the both soultions and the centroid are compared.
    The new position with the z-coordinate above the centroid is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update1w2 = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the first particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Center coordinates of the first auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Radius of the first auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Coordinates of the wall
    z_w1 = z1
    z_w2 = z1
    if walls == 3:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == -3:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == 4:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    elif walls == -4:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    """ Calculation of the triangle's plane """
    # Calculate vectors AB and AC
    AB = np.array([z_w2 - z1, y_w2 - y1, x_w2 - x1])
    AC = np.array([z_w1 - z1, y_w1 - y1, x_w1 - x1])
    # Compute the normal vector to the plane (cross product of AB and AC)
    V_triangle_norm = np.cross(AB, AC)
    A, B, C = V_triangle_norm
    # Calculate D using the plane equation Ax + By + Cz + D = 0
    # Using point A (x1, y1, z1) to find D
    D = -(A * z1 + B * y1 + C * x1)
    """ 
    Calculation of the reflected position of the considered particle.
    First, the distance between the center of the considered particle and the triangle's plane is determined.
    Then, the position of the considered particle is reflected about the triangle's plane
    The coordiantes of the reflected position are used as second guess for the solver to determine the second solver solution. 
    """
    # Calculate the distance from the considered particle's center to the plane
    distance_pi = (A * z_i + B * y_i + C * x_i + D) / np.linalg.norm(V_triangle_norm)
    # Calculate the reflection point on the plane
    P_reflected = np.array([z_i, y_i, x_i]) - 2 * distance_pi * V_triangle_norm / np.linalg.norm(V_triangle_norm)
    """ Solver to find possible new position of the considered particle """
    # Coordinates for the initial guess
    z_01 = z_i
    z_02 = P_reflected[0]
    # Initial guess values
    initial_guess1 = z_01
    initial_guess2 = z_02
    # X and y-coordinates of the new position
    if walls == 3:
        x = config.A_0 - d_i / 2
        y = config.A_0 - d_i / 2
    elif walls == -3:
        x = -config.A_0 + d_i / 2 
        y = config.A_0 - d_i / 2
    elif walls == 4:
        x = config.A_0 - d_i / 2
        y = -config.A_0 + d_i / 2
    elif walls == -4:
        x = -config.A_0 + d_i / 2
        y = -config.A_0 + d_i / 2
    # Solver
    z_s1 = fsolve(solver_position1w2, initial_guess1, args=(c1, r1, x, y))
    z_s2 = fsolve(solver_position1w2, initial_guess2, args=(c1, r1, x, y))
    z_s1 = z_s1[0]
    z_s2 = z_s2[0]  
    """ Check wheter the solver found two different solutions or the same twice """
    # Error value is 1 if the solver found the same position twice
    if z_s1 == z_s2:
        error_position_update1w2 = 1
    """ 
    Determine the actual new position of the considered particle from the both solver solutions.
    Must be the closest position to the original position of the considered particle
    """
    # Z-coordinate of polygon's centroid    
    z_cen = z1
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the first solver solution are assigned to the center of the considered particle
    if z_s1 > z_cen:
        z = z_s1
    # If the center of the considered particle is located above the centroid, 
    # the coordinates of the second solver solution are assigned to the center of the considered particle
    elif z_s2 > z_cen: 
        z = z_s2
    else: 
        error_position_update1w2 = 2
        return z_i, y_i, x_i, error_position_update1w2
    """ 
    Check wheter the distances between the centers of the three steady state particles 
    and the new center position of the considered particle are in the samerange as the sums of their radii.
    """
    same_range = check_distances1w2(r1, c1, d_i, z, y, x)
    # Error value is 2 if the distance between the steady state particles and the position of the solver solution 
    # is not in the same range as the sums of their radii
    if same_range == False:
        error_position_update1w2 = 3
        return z_i, y_i, x_i, error_position_update1w2

    return z, y, x, error_position_update1w2

def solver_position1w2(fit_pars, c1, r1, x_s, y_s):
    """ Solver to calculate the new position of the considered particle if it overlaps one particle and two walls """
    # Solver variables
    z = fit_pars
    # Equation of a circle. The circles coresponds here to the auxiliary spheres
    eq1 = (x_s - c1[2])**2 + (y_s - c1[1])**2 + (z - c1[0])**2 - r1**2
    return eq1

def check_distances1w2(config, r1, c1, d_i, z_s, y_s, x_s): 
    """ Function that checks wheter the distances between the center of one steady state particles, the walls, 
    and the new center position of the considered particle are in the same range as the sums of their radii. """
    # True if distances are in the same range as the sums of radii
    same_range = True
    # Vectors between the centers of the steady state particles and the new center positions of the considered particle    
    V_centers_1i = np.array([z_s, y_s, x_s]) - c1
    # Distance between the centers of the steady state particles and the new center position of the considered particle
    dist_centers_1i = np.linalg.norm(V_centers_1i)
    # Distances between the center of the considered particle and the walls
    dist_w1i = config.A_0 - abs(x_s) 
    dist_w2i = config.A_0 - abs(y_s)
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    if abs(dist_centers_1i-r1) > 1e-9 or abs(dist_w2i-d_i/2) > 1e-9 or abs(dist_w1i-d_i/2) > 1e-9:
        same_range = False

    return same_range

def check_steadystate1w2(config, z_i, y_i, x_i, M_particle_steady, overlaps, walls):
    """ 
    Function to check the steady state of the considered particle if it is in contact with one steady state particle and two walls.
    As soon as considered particle collides with one steady state particle and two walls, a steady state check is performed.
    Considered particle is in steady state if its center lies within the y-x-plane of a triangle located between the 
    steady state particle center and the walls. 
    Barycentric coordinate method is used to determine if the center of the considered particle lies within the polygon.
    Moreover, barycentric coordinates are used to determine the subdomain of the triangle where the center lies.
    Based on the subdomain, the particle rolling direction is identified.
    """
    # Used in case an error occurs during steadystate check.
    error_check_steadystate1w2 = 0
    # Variable for steady state check condition
    steady_check = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Coordinates of the wall
    if walls == 3:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == -3:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == 4:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    elif walls == -4:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    # Z-coordinate of triangle's centroid
    z_cen = z1
    # Check whether the center of the considered particle is located above the centroid
    # Double checking at this point, since already handled in the position_update3 function.
    # Error value is 1 when the center is located below the centroid
    if z_i < z_cen:
        error_check_steadystate1w2 = 1
        return steady_check, overlaps, error_check_steadystate1w2
    """ Barycentric coordinate technique """
    # Denominator
    denominator = ((y_w2 - y_w1) * (x1 - x_w1) + (x_w1 - x_w2) * (y1 - y_w1))
    # Check whether the triangle points are collinear
    # Error value is 2 when the triangle points are collinear
    if denominator == 0:
        error_check_steadystate1w2 = 2  
        return steady_check, overlaps, error_check_steadystate1w2
    # Barycentric coordinates
    a = ((y_w2 - y_w1) * (x_i - x_w1) + (x_w1 - x_w2) * (y_i - y_w1)) / denominator
    b = ((y_w1 - y1) * (x_i - x_w1) + (x1 - x_w1) * (y_i - y_w1)) / denominator
    c = 1 - a - b
    # Determine the subdomain of the triangle and thus wheter it is in steady state or in which direction it is rolling
    if (0 <= a <= 1) and (0 <= b <= 1) and (0 <= c <= 1):
        steady_check = 3
    elif b < 0 and a >= 0 and c >= 0:
        # Rolling along steady state particle Nr.1 and wall Nr.2
        if walls == 3 or walls == -3:
            walls = 2
        elif walls == 4 or walls == -4:
            walls = -2
        steady_check = 4
    elif c < 0 and a >= 0 and b >= 0:
        # Rolling along steady state particle Nr.1 and wall Nr.1
        if walls == 3 or walls == 4:
            walls = 1
        elif walls == -3 or walls == -4:
            walls = -1
        steady_check = 4
    elif b < 0 and c < 0 and a > 1:
        # Rolling along steady state particle Nr.1 
        steady_check = 1
    else:
        # Error value is 3 when the subdomain of the triangle could not be determined correctly
        error_check_steadystate1w2 = 3

    return steady_check, overlaps, walls, error_check_steadystate1w2

def position_update2w2(config, z_i, y_i, x_i, d_i, M_particle_steady, overlaps, walls):
    """
    Function to update the position of the considered particle if it overlaps with two particles and two walls.
    There are two mathematical solutions of the new position, but only one logical. 
    A solver is used to determine these solutions.
    First, a polygon plane is calculated which crosses the centers of the two steady state particles and the walls.
    Moreover, the centroid of the polygon is determined.
    The polygon plane is used to identify the guesses of the both solutions of the considered particle's new position.
    It is tested whether the both solutions differ and if they are correct.
    To identify the logical solution of the both solutions, the z-coordinates of the both soultions and the centroid are compared.
    The new position with the z-coordinate above the centroid is the logical solution.
    """
    # Used in case an error occurs during position update.
    error_position_update2w2 = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Diameter of the first particle in steady state
    d1 = float(M_particle_steady[overlaps[0], 5])
    # Radius of the first auxiliary sphere 
    r1 = (d_i + d1) / 2
    # Center coordinates of the first auxiliary sphere
    c1 = np.array([z1, y1, x1])
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Diameter of the second particle in steady state
    d2 = float(M_particle_steady[overlaps[1], 5])
    # Radius of the first auxiliary sphere 
    r2 = (d_i + d2) / 2
    # Center coordinates of the second auxiliary sphere
    c2 = np.array([z2, y2, x2])
    # Coordinates of the wall
    z_w1 = (z1+z2)/2
    z_w2 = (z1+z2)/2
    """ Solver to find possible new position of the considered particle """    
    # Array with the center coordinates and radii of the steady state particles
    particles = [
    {'center': c1, 'diameter': d1},
    {'center': c2, 'diameter': d2}]
    # Coordinates for the initial guess
    x_01 = x_i
    y_01 = y_i
    z_01 = z_i
    # Initial guess values
    initial_guess = [x_01, y_01, z_01]
    # Bounds
    bnds = [(None, None), (-config.A_0, config.A_0), (-config.A_0, config.A_0)]
    # Perform the optimization
    result = minimize(solver_position2w2, initial_guess, args=(config, particles, d_i), method='L-BFGS-B', bounds=bnds, tol=1e-10)
    # Extract the coordinates
    z_s, y_s, x_s = result.x[:3]
    """ 
    Check wheter the distances between the centers of the four steady state particles 
    and the new center position of the considered particle are in the samerange as the sums of their radii.
    """
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    same_range = check_distances2w2(config, r1, r2, c1, c2, d_i, z_s, y_s, x_s)
    # Error value is 1 if the distance between the steay state particles and the position of the first solver solution 
    # is not in the same range as the sums of their radii
    if same_range == False:
        error_position_update2w2 = 1
        return z_i, y_i, x_i, error_position_update2w2
    """ 
    Check whether the new position of the considered particles is located above the centroid of the steady state particles
    """
    # Z-coordinate of polygon's centroid
    z_cen = (z1 + z2 + z_w1 + z_w2) / 4
    # If the center of the considered particle is located above the centroid, 
    # the new coordinates are assigned to the center of the considered particle
    if z_s > z_cen:
        z = z_s
        y = y_s
        x = x_s   
    # If the center of the considered particle is located below the centroid, 
    # the new coordinates are not correct 
    else:
        error_position_update2w2 = 2
        return z_i, y_i, x_i, error_position_update2w2

    return z, y, x, error_position_update2w2

def solver_position2w2(fit_pars, config, particles, d_i):
    """ Solver to calculate the new position of the considered particle if it overlaps with two particles and two walls """
    # Center coordinates of the considered particle are the optimizing parameters
    c_i = np.array(fit_pars[:3])
    # Mini
    total_error = 0
    # Calculate the total error between the actual and required distances of the particle centers
    for particle in particles:
        # Center of the steady state particle
        c_j = particle['center']
        # Diameter of the steady state particle
        d_j = particle['diameter']
        # Actual distance between the centers of the considered particle and the steady state particle
        distance = np.linalg.norm(c_i - c_j)
        # Required distance between the centers of the considered particle and the steady state particle
        required_distance = (d_i + d_j)/2
        # Minimization function
        total_error += 100*(distance - required_distance)/required_distance
    # Actual distance between the center of the considered particle and the wall
    distance_w1 = config.A_0 - abs(c_i[2])
    distance_w2 = config.A_0 - abs(c_i[1])
    # Required distance between the center of the considered particle and the steady state particle
    required_distance_w = d_i/2
    # Update minimization function
    total_error += 100*abs(distance_w1 - required_distance_w)/required_distance_w
    total_error += 100*abs(distance_w2 - required_distance_w)/required_distance_w

    return total_error

def check_distances2w2(config, r1, r2, c1, c2, d_i, z_s, y_s, x_s): 
    """ Function that checks wheter the distances between the centers of the two steady state particles, the walls, 
    and the new center position of the considered particle are in the same range as the sums of their radii. """
    # True if distances are in the same range as the sums of radii
    same_range = True
    # Vectors between the centers of the steady state particles and the new center positions of the considered particle    
    V_centers_1i = np.array([z_s, y_s, x_s]) - c1
    V_centers_2i = np.array([z_s, y_s, x_s]) - c2
    # Distance between the centers of the steady state particles and the new center position of the considered particle
    dist_centers_1i = np.linalg.norm(V_centers_1i)
    dist_centers_2i = np.linalg.norm(V_centers_2i)
    # Distance between the center of the considered particle and the wall 
    dist_w1i = config.A_0 - abs(x_s)
    dist_w2i = config.A_0 - abs(y_s)
    # Check if the left-hand side is approximately zero (account for floating-point precision)
    if abs(dist_centers_1i-r1) > 1e-9 or abs(dist_centers_2i-r2) > 1e-9 or abs(dist_w1i-d_i/2) > 1e-9 or abs(dist_w2i-d_i/2) > 1e-9:
        same_range = False

    return same_range

def check_steadystate2w2(config, z_i, y_i, x_i, M_particle_steady, overlaps, walls):
    """ 
    Function to check the steady state of the considered particle if it is in contact with two steady state particles and two walls.
    As soon as considered particle collides with two steady state particles and two walls, a steady state check is performed.
    Considered particle is in steady state if its center lies within the y-x-plane of a polygon located between the 
    steady state particle centers and the walls. 
    Barycentric coordinate method is used to determine if the center of the condsidered particle lies within the polygon.
    For this purpose, the polygon is devided into two triangles.
    Moreover, barycentric coordinates are used to determine the subdomain of the polygon where the point lies.
    Based on the subdomain, the particle rolling direction is identified.
    """
    # Used in case an error occurs during steadystate check.
    error_check_steadystate2w2 = 0
    # Variable for steady state check condition
    steady_check = 0
    # Coordinates of the first particle in steady state
    z1 = float(M_particle_steady[overlaps[0], 1])
    y1 = float(M_particle_steady[overlaps[0], 2])
    x1 = float(M_particle_steady[overlaps[0], 3])
    # Coordinates of the second particle in steady state
    z2 = float(M_particle_steady[overlaps[1], 1])
    y2 = float(M_particle_steady[overlaps[1], 2])
    x2 = float(M_particle_steady[overlaps[1], 3])
    # Coordinates of the wall
    z_w1 = (z1+z2)/2
    z_w2 = (z1+z2)/2
    if walls == 3:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == -3:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = config.A_0
    elif walls == 4:
        x_w1 = config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    elif walls == -4:
        x_w1 = -config.A_0
        y_w1 = y_i
        x_w2 = x_i
        y_w2 = -config.A_0
    # Z-coordinate of polygon's centroid
    z_cen = (z1 + z2 + z_w1 + z_w2) / 4
    # Check whether the center of the considered particle is located above the centroid
    # Double checking at this point, since already handled in the position_update4 function.
    # Error value is 1 when the center is located below the centroid
    if z_i < z_cen:
        error_check_steadystate2w2 = 1
        return steady_check, overlaps, error_check_steadystate2w2
    """ 
    The polygon is devided into two triangles. 
    The first triangle is defined by the centers of the steady state particles 1, 2 and 4. 1=A, 2=B, 4=D
    The second triangle is defined by the centers of the steady state particles 2, 3 and 4. 2=B, 3=C, 4=D
    Barycentric coordinate technique is applied on each triangle
    """
    # Denominator of the first triangle
    denominator1 = ((y2 - y_w1) * (x1 - x_w1) + (x_w1 - x2) * (y1 - y_w1))
    # Denominator of the second triangle
    denominator2 = ((y_w2 - y_w1) * (x2 - x_w1) + (x_w1 - x_w2) * (y2 - y_w1))
    # Check whether the triangle points are collinear
    # Error value is 2 when the triangle points are collinear
    if denominator1 == 0 or denominator2 == 0:
        error_check_steadystate2w2 = 2 
        return steady_check, overlaps, error_check_steadystate2w2 
    # Barycentric coordinates of the first triangle
    a = ((y2 - y_w1) * (x_i - x_w1) + (x_w1 - x2) * (y_i - y_w1)) / denominator1
    b1 = ((y_w1 - y1) * (x_i - x_w1) + (x1 - x_w1) * (y_i - y_w1)) / denominator1
    d1 = 1 - a - b1
    # Barycentric coordinates of the second triangle
    b2 = ((y_w2 - y_w1) * (x_i - x_w1) + (x_w1 - x_w2) * (y_i - y_w1)) / denominator2
    c = ((y_w1 - y2) * (x_i - x_w1) + (x2 - x_w1) * (y_i - y_w1)) / denominator2
    d2 = 1 - b2 - c
    # Determine the subdomain of the polygon and thus wheter it is in steady state or in which direction it is rolling
    if (0 <= a <= 1) and (0 <= b1 <= 1) and (0 <= d1 <= 1):
        steady_check = 3
    elif (0 <= b2 <= 1) and (0 <= c <= 1) and (0 <= d2 <= 1):
        steady_check = 3
    elif b1 < 0 and d1 < 0:
        # Rolling along steady state particle Nr.1
        overlaps = np.delete(overlaps, 1)
        steady_check = 1
    elif d1 < 0 and d2 < 0:
        # Rolling along steady state particle Nr.2
        overlaps = np.delete(overlaps, 0)
        steady_check = 1
    elif d1 < 0 and b1 >= 0 and d2 >= 0:
        # Rolling along steady state particle Nr.1 and Nr.2
        steady_check = 2
    elif d2 < 0 and b2 >= 0 and d1 >= 0:
        # Rolling along steady state particle Nr.2 and wall Nr.1
        if walls == 3 or walls == 4:
            walls = 1
        elif walls == -3 or walls == -4:
            walls = -1
        overlaps = np.delete(overlaps, 0)
        steady_check = 4
    elif b1 < 0 and d1 >= 0 and b2 >= 0:
        # Rolling along steady state particle Nr.1 and wall Nr.2
        if walls == 3 or walls == -3:
            walls = 2
        elif walls == 4 or walls == -4:
            walls = -2
        overlaps = np.delete(overlaps, 1)
        steady_check = 4
    else:
        # Error value is 3 when the subdomain of the polygon could not be determined correctly
        error_check_steadystate2w2 = 3

    return steady_check, overlaps, walls, error_check_steadystate2w2